# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

# Copyright (c) 2023, Tri Dao
# https://github.com/state-spaces/mamba/blob/fb7b5310fa865dbd62aa059b1e26f2b431363e2a/mamba_ssm/ops/triton/layernorm.py
# Implement residual + layer_norm / rms_norm.

# Based on the Triton LayerNorm tutorial: https://triton-lang.org/main/getting-started/tutorials/05-layer-norm.html
# For the backward pass, we keep weight_grad and bias_grad in registers and accumulate.
# This is faster for dimensions up to 8k, but after that it's much slower due to register spilling.
# The models we train have hidden dim up to 8k anyway (e.g. Llama 70B), so this is fine.

from __future__ import annotations

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
import triton
import triton.language as tl
from einops import rearrange
from torch.distributed import DeviceMesh
from torch.distributed.tensor import Replicate, Shard, distribute_module
from torch.distributed.tensor.parallel import ParallelStyle

from fla.utils import autotune_cache_kwargs, get_multiprocessor_count, input_guard

try:
    from torch.distributed.tensor import DTensor
except (ImportError, AttributeError):
    DTensor = None


def layer_norm_ref(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    upcast: bool = False,
):
    dtype = x.dtype
    if upcast:
        weight = weight.float()
        bias = bias.float() if bias is not None else None
    if upcast:
        x = x.float()
        residual = residual.float() if residual is not None else residual
    if residual is not None:
        x = (x + residual).to(x.dtype)
    out = F.layer_norm(x.to(weight.dtype), x.shape[-1:], weight=weight, bias=bias, eps=eps).to(
        dtype,
    )
    return out if not prenorm else (out, x)


def rms_norm_ref(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    upcast: bool = False,
):
    dtype = x.dtype
    if upcast:
        weight = weight.float()
        bias = bias.float() if bias is not None else None
    if upcast:
        x = x.float()
        residual = residual.float() if residual is not None else residual
    if residual is not None:
        x = (x + residual).to(x.dtype)
    rstd = 1 / torch.sqrt((x.square()).mean(dim=-1, keepdim=True) + eps)
    out = (x * rstd * weight) + bias if bias is not None else (x * rstd * weight)
    out = out.to(dtype)
    return out if not prenorm else (out, x)


def group_norm_ref(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    num_groups: int,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    is_rms_norm: bool = False,
    prenorm: bool = False,
    upcast: bool = False,
):
    dtype = x.dtype
    if upcast:
        weight = weight.float()
        bias = bias.float() if bias is not None else None
    if upcast:
        x = x.float()
        residual = residual.float() if residual is not None else residual
    if residual is not None:
        x = (x + residual).to(x.dtype)
    residual = x
    x, weight = [
        rearrange(data, "... (g d) -> ... g d", g=num_groups) for data in (x, weight)
    ]
    if bias is not None:
        bias = rearrange(bias, '... (g d) -> ... g d', g=num_groups)
    if not is_rms_norm:
        mean = x.mean(dim=-1, keepdim=True)
        x = x - mean
    rstd = 1 / torch.sqrt((x.square()).mean(dim=-1, keepdim=True) + eps)
    out = (x * rstd * weight) + bias if bias is not None else (x * rstd * weight)
    out = rearrange(out, "... g d -> ... (g d)")
    out = out.to(dtype)
    return out if not prenorm else (out, residual)


class GroupNormRef(nn.Module):

    def __init__(
        self,
        num_groups: int,
        hidden_size: int,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        is_rms_norm: bool = False,
    ) -> GroupNormRef:
        super().__init__()

        if hidden_size % num_groups != 0:
            raise ValueError('num_channels must be divisible by num_groups')

        self.num_groups = num_groups
        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps
        self.is_rms_norm = is_rms_norm

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.num_groups}, {self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        if self.is_rms_norm:
            s += f", is_rms_norm={self.is_rms_norm}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, residual=None, prenorm=False):
        return group_norm_ref(
            x,
            self.weight,
            self.bias,
            num_groups=self.num_groups,
            residual=residual,
            eps=self.eps,
            is_rms_norm=self.is_rms_norm,
            prenorm=prenorm,
            upcast=True,
        )


@triton.autotune(
    configs=[
        triton.Config({'BT': BT}, num_warps=num_warps)
        for BT in [32, 64, 128]
        for num_warps in [2, 4, 8]
    ],
    key=['D', 'NB', 'HAS_RESIDUAL', 'STORE_RESIDUAL_OUT', 'IS_RMS_NORM'],
    **autotune_cache_kwargs,
)
@triton.jit
def layer_norm_fwd_kernel(
    x,  # pointer to the input
    y,  # pointer to the output
    w,  # pointer to the weights
    b,  # pointer to the biases
    res,  # pointer to the res
    res_out,  # pointer to the res
    mean,  # pointer to the mean
    rstd,  # pointer to the 1/std
    eps,  # epsilon to avoid division by zero
    T,
    G: tl.constexpr,
    D: tl.constexpr,
    BT: tl.constexpr,
    BD: tl.constexpr,
    NB: tl.constexpr,
    IS_RMS_NORM: tl.constexpr,
    HAS_RESIDUAL: tl.constexpr,
    STORE_RESIDUAL_OUT: tl.constexpr,
    HAS_WEIGHT: tl.constexpr,
    HAS_BIAS: tl.constexpr,
):
    i_t = tl.program_id(0)

    o_t = i_t * BT + tl.arange(0, BT)
    o_g = o_t % G
    o_d = tl.arange(0, BD)
    m_d = o_d < D

    p_x = tl.make_block_ptr(x, (T, D), (D, 1), (i_t * BT, 0), (BT, BD), (1, 0))
    b_x = tl.load(p_x, boundary_check=(0, 1)).to(tl.float32)
    if HAS_RESIDUAL:
        p_res = tl.make_block_ptr(res, (T, D), (D, 1), (i_t * BT, 0), (BT, BD), (1, 0))
        b_x += tl.load(p_res, boundary_check=(0, 1)).to(tl.float32)
    if STORE_RESIDUAL_OUT:
        p_res_out = tl.make_block_ptr(res_out, (T, D), (D, 1), (i_t * BT, 0), (BT, BD), (1, 0))
        tl.store(p_res_out, b_x.to(p_res_out.dtype.element_ty), boundary_check=(0, 1))
    if not IS_RMS_NORM:
        b_mean = tl.sum(b_x, axis=1) / D
        p_mean = tl.make_block_ptr(mean, (T,), (1,), (i_t * BT,), (BT,), (0,))
        tl.store(p_mean, b_mean.to(p_mean.dtype.element_ty), boundary_check=(0,))
        b_xbar = tl.where(m_d[None, :], b_x - b_mean[:, None], 0.0)
        b_var = tl.sum(b_xbar * b_xbar, axis=1) / D
    else:
        b_xbar = tl.where(m_d[None, :], b_x, 0.0)
        b_var = tl.sum(b_xbar * b_xbar, axis=1) / D
    b_rstd = 1 / tl.sqrt(b_var + eps)

    p_rstd = tl.make_block_ptr(rstd, (T,), (1,), (i_t * BT,), (BT,), (0,))
    tl.store(p_rstd, b_rstd.to(p_rstd.dtype.element_ty), boundary_check=(0,))

    if HAS_WEIGHT:
        b_w = tl.load(w + o_g[:, None] * D + o_d[None, :], mask=m_d[None, :]).to(tl.float32)
    if HAS_BIAS:
        b_b = tl.load(b + o_g[:, None] * D + o_d[None, :], mask=m_d[None, :]).to(tl.float32)
    b_x_hat = (b_x - b_mean[:, None]) * b_rstd[:, None] if not IS_RMS_NORM else b_x * b_rstd[:, None]
    b_y = b_x_hat * b_w if HAS_WEIGHT else b_x_hat
    if HAS_BIAS:
        b_y = b_y + b_b

    # Write output
    p_y = tl.make_block_ptr(y, (T, D), (D, 1), (i_t * BT, 0), (BT, BD), (1, 0))
    tl.store(p_y, b_y.to(p_y.dtype.element_ty), boundary_check=(0, 1))


@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [2, 4, 8, 16]
    ],
    key=['D', 'HAS_RESIDUAL', 'STORE_RESIDUAL_OUT', 'IS_RMS_NORM'],
    **autotune_cache_kwargs,
)
@triton.jit
def layer_norm_fwd_kernel1(
    x,  # pointer to the input
    y,  # pointer to the output
    w,  # pointer to the weights
    b,  # pointer to the biases
    res,  # pointer to the res
    res_out,  # pointer to the res
    mean,  # pointer to the mean
    rstd,  # pointer to the 1/std
    eps,  # epsilon to avoid division by zero
    G: tl.constexpr,
    D: tl.constexpr,
    BD: tl.constexpr,
    IS_RMS_NORM: tl.constexpr,
    HAS_RESIDUAL: tl.constexpr,
    STORE_RESIDUAL_OUT: tl.constexpr,
    HAS_WEIGHT: tl.constexpr,
    HAS_BIAS: tl.constexpr,
):
    i_t = tl.program_id(0)
    i_g = i_t % G

    x += i_t * D
    y += i_t * D
    if HAS_RESIDUAL:
        res += i_t * D
    if STORE_RESIDUAL_OUT:
        res_out += i_t * D

    o_d = tl.arange(0, BD)
    m_d = o_d < D
    b_x = tl.load(x + o_d, mask=m_d, other=0.0).to(tl.float32)
    if HAS_RESIDUAL:
        b_x += tl.load(res + o_d, mask=m_d, other=0.0).to(tl.float32)
    if STORE_RESIDUAL_OUT:
        tl.store(res_out + o_d, b_x, mask=m_d)
    if not IS_RMS_NORM:
        b_mean = tl.sum(b_x, axis=0) / D
        tl.store(mean + i_t, b_mean)
        b_xbar = tl.where(m_d, b_x - b_mean, 0.0)
        b_var = tl.sum(b_xbar * b_xbar, axis=0) / D
    else:
        b_xbar = tl.where(m_d, b_x, 0.0)
        b_var = tl.sum(b_xbar * b_xbar, axis=0) / D
    b_rstd = 1 / tl.sqrt(b_var + eps)
    tl.store(rstd + i_t, b_rstd)

    if HAS_WEIGHT:
        b_w = tl.load(w + i_g * D + o_d, mask=m_d).to(tl.float32)
    if HAS_BIAS:
        b_b = tl.load(b + i_g * D + o_d, mask=m_d).to(tl.float32)
    b_x_hat = (b_x - b_mean) * b_rstd if not IS_RMS_NORM else b_x * b_rstd
    b_y = b_x_hat * b_w if HAS_WEIGHT else b_x_hat
    if HAS_BIAS:
        b_y = b_y + b_b

    # Write output
    tl.store(y + o_d, b_y, mask=m_d)


@triton.heuristics({
    'RECOMPUTE_OUTPUT': lambda args: args['y'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BT': BT}, num_warps=num_warps)
        for BT in [32, 64]
        for num_warps in [2, 4, 8]
    ],
    key=['D', 'NB', 'HAS_DRESIDUAL', 'STORE_DRESIDUAL', 'IS_RMS_NORM'],
    **autotune_cache_kwargs,
)
@triton.jit
def layer_norm_bwd_kernel(
    x,  # pointer to the input
    w,  # pointer to the weights
    b,  # pointer to the biases
    y,  # pointer to the output to be recomputed
    dy,  # pointer to the output gradient
    dx,  # pointer to the input gradient
    dw,  # pointer to the partial sum of weights gradient
    db,  # pointer to the partial sum of biases gradient
    dres,
    dres_in,
    mean,
    rstd,
    T,
    G: tl.constexpr,
    D: tl.constexpr,
    BS: tl.constexpr,
    BT: tl.constexpr,
    BD: tl.constexpr,
    NB: tl.constexpr,
    GS: tl.constexpr,
    IS_RMS_NORM: tl.constexpr,
    HAS_DRESIDUAL: tl.constexpr,
    STORE_DRESIDUAL: tl.constexpr,
    HAS_WEIGHT: tl.constexpr,
    HAS_BIAS: tl.constexpr,
    RECOMPUTE_OUTPUT: tl.constexpr,
):
    i_s = tl.program_id(0)
    i_g, i_sg = i_s // GS, i_s % GS

    o_d = tl.arange(0, BD)
    m_d = o_d < D
    if HAS_WEIGHT:
        b_w = tl.load(w + i_g * D + o_d, mask=m_d).to(tl.float32)
        b_dw = tl.zeros((BT, BD), dtype=tl.float32)
    if HAS_BIAS:
        b_b = tl.load(b + i_g * D + o_d, mask=m_d, other=0.0).to(tl.float32)
        b_db = tl.zeros((BT, BD), dtype=tl.float32)

    T = min(i_sg * BS + BS, T // G)
    for i_t in range(i_sg * BS, T, BT):
        p_x = tl.make_block_ptr(x + i_g * D, (T, D), (G*D, 1), (i_t, 0), (BT, BD), (1, 0))
        p_dy = tl.make_block_ptr(dy + i_g * D, (T, D), (G*D, 1), (i_t, 0), (BT, BD), (1, 0))
        p_dx = tl.make_block_ptr(dx + i_g * D, (T, D), (G*D, 1), (i_t, 0), (BT, BD), (1, 0))
        # [BT, BD]
        b_x = tl.load(p_x, boundary_check=(0, 1)).to(tl.float32)
        b_dy = tl.load(p_dy, boundary_check=(0, 1)).to(tl.float32)

        if not IS_RMS_NORM:
            p_mean = tl.make_block_ptr(mean + i_g, (T,), (G,), (i_t,), (BT,), (0,))
            b_mean = tl.load(p_mean, boundary_check=(0,))
        p_rstd = tl.make_block_ptr(rstd + i_g, (T,), (G,), (i_t,), (BT,), (0,))
        b_rstd = tl.load(p_rstd, boundary_check=(0,))
        # Compute dx
        b_xhat = (b_x - b_mean[:, None]) * b_rstd[:, None] if not IS_RMS_NORM else b_x * b_rstd[:, None]
        b_xhat = tl.where(m_d[None, :], b_xhat, 0.0)

        b_y = b_xhat * b_w[None, :] if HAS_WEIGHT else b_xhat
        if HAS_BIAS:
            b_y = b_y + b_b[None, :]
        if RECOMPUTE_OUTPUT:
            p_y = tl.make_block_ptr(y + i_g * D, (T, D), (G*D, 1), (i_t, 0), (BT, BD), (1, 0))
            tl.store(p_y, b_y.to(p_y.dtype.element_ty), boundary_check=(0, 1))

        b_wdy = b_dy

        if HAS_WEIGHT or HAS_BIAS:
            m_t = (i_t + tl.arange(0, BT)) < T
        if HAS_WEIGHT:
            b_wdy = b_dy * b_w
            b_dw += tl.where(m_t[:, None], b_dy * b_xhat, 0.0)
        if HAS_BIAS:
            b_db += tl.where(m_t[:, None], b_dy, 0.0)
        if not IS_RMS_NORM:
            b_c1 = tl.sum(b_xhat * b_wdy, axis=1) / D
            b_c2 = tl.sum(b_wdy, axis=1) / D
            b_dx = (b_wdy - (b_xhat * b_c1[:, None] + b_c2[:, None])) * b_rstd[:, None]
        else:
            b_c1 = tl.sum(b_xhat * b_wdy, axis=1) / D
            b_dx = (b_wdy - b_xhat * b_c1[:, None]) * b_rstd[:, None]
        if HAS_DRESIDUAL:
            p_dres = tl.make_block_ptr(dres + i_g * D, (T, D), (G*D, 1), (i_t, 0), (BT, BD), (1, 0))
            b_dres = tl.load(p_dres, boundary_check=(0, 1)).to(tl.float32)
            b_dx += b_dres
        # Write dx
        if STORE_DRESIDUAL:
            p_dres_in = tl.make_block_ptr(dres_in + i_g * D, (T, D), (G*D, 1), (i_t, 0), (BT, BD), (1, 0))
            tl.store(p_dres_in, b_dx.to(p_dres_in.dtype.element_ty), boundary_check=(0, 1))

        tl.store(p_dx, b_dx.to(p_dx.dtype.element_ty), boundary_check=(0, 1))

    if HAS_WEIGHT:
        tl.store(dw + i_s * D + o_d, tl.sum(b_dw, axis=0), mask=m_d)
    if HAS_BIAS:
        tl.store(db + i_s * D + o_d, tl.sum(b_db, axis=0), mask=m_d)


@triton.heuristics({
    'RECOMPUTE_OUTPUT': lambda args: args['y'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [2, 4, 8]
    ],
    key=['D', 'HAS_DRESIDUAL', 'STORE_DRESIDUAL', 'IS_RMS_NORM'],
    **autotune_cache_kwargs,
)
@triton.jit
def layer_norm_bwd_kernel1(
    x,  # pointer to the input
    w,  # pointer to the weights
    b,  # pointer to the biases
    y,  # pointer to the output to be recomputed
    dy,  # pointer to the output gradient
    dx,  # pointer to the input gradient
    dw,  # pointer to the partial sum of weights gradient
    db,  # pointer to the partial sum of biases gradient
    dres,
    dres_in,
    mean,
    rstd,
    T,
    G: tl.constexpr,
    D: tl.constexpr,
    BS: tl.constexpr,
    BD: tl.constexpr,
    GS: tl.constexpr,
    IS_RMS_NORM: tl.constexpr,
    HAS_DRESIDUAL: tl.constexpr,
    STORE_DRESIDUAL: tl.constexpr,
    HAS_WEIGHT: tl.constexpr,
    HAS_BIAS: tl.constexpr,
    RECOMPUTE_OUTPUT: tl.constexpr,
):
    i_s = tl.program_id(0)
    i_g, i_sg = i_s // GS, i_s % GS

    o_d = tl.arange(0, BD)
    mask = o_d < D

    if HAS_WEIGHT:
        b_w = tl.load(w + i_g * D + o_d, mask=mask).to(tl.float32)
        b_dw = tl.zeros((BD,), dtype=tl.float32)
    if RECOMPUTE_OUTPUT and HAS_BIAS:
        b_b = tl.load(b + i_g * D + o_d, mask=mask, other=0.0).to(tl.float32)
    if HAS_BIAS:
        b_db = tl.zeros((BD,), dtype=tl.float32)

    for i_t in range(i_sg * BS * G + i_g, min((i_sg * BS + BS) * G + i_g, T), G):
        b_x = tl.load(x + i_t * D + o_d, mask=mask, other=0).to(tl.float32)
        b_dy = tl.load(dy + i_t * D + o_d, mask=mask, other=0).to(tl.float32)

        if not IS_RMS_NORM:
            b_mean = tl.load(mean + i_t)
        b_rstd = tl.load(rstd + i_t)
        # Compute dx
        b_xhat = (b_x - b_mean) * b_rstd if not IS_RMS_NORM else b_x * b_rstd
        b_xhat = tl.where(mask, b_xhat, 0.0)
        if RECOMPUTE_OUTPUT:
            b_y = b_xhat * b_w if HAS_WEIGHT else b_xhat
            if HAS_BIAS:
                b_y = b_y + b_b
            tl.store(y + i_t * D + o_d, b_y, mask=mask)
        b_wdy = b_dy
        if HAS_WEIGHT:
            b_wdy = b_dy * b_w
            b_dw += b_dy * b_xhat
        if HAS_BIAS:
            b_db += b_dy
        if not IS_RMS_NORM:
            b_c1 = tl.sum(b_xhat * b_wdy, axis=0) / D
            b_c2 = tl.sum(b_wdy, axis=0) / D
            b_dx = (b_wdy - (b_xhat * b_c1 + b_c2)) * b_rstd
        else:
            b_c1 = tl.sum(b_xhat * b_wdy, axis=0) / D
            b_dx = (b_wdy - b_xhat * b_c1) * b_rstd
        if HAS_DRESIDUAL:
            b_dres = tl.load(dres + i_t * D + o_d, mask=mask, other=0).to(tl.float32)
            b_dx += b_dres
        # Write dx
        b_dx = tl.cast(b_dx, dtype=dx.dtype.element_ty, fp_downcast_rounding='rtne')
        if STORE_DRESIDUAL:
            tl.store(dres_in + i_t * D + o_d, b_dx, mask=mask)
        tl.store(dx + i_t * D + o_d, b_dx, mask=mask)

    if HAS_WEIGHT:
        tl.store(dw + i_s * D + o_d, b_dw, mask=mask)
    if HAS_BIAS:
        tl.store(db + i_s * D + o_d, b_db, mask=mask)


def layer_norm_fwd(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    eps: float = 1e-5,
    residual: torch.Tensor = None,
    out_dtype: torch.dtype = None,
    residual_dtype: torch.dtype = None,
    is_rms_norm: bool = False,
    num_groups: int = 1,
):
    if residual is not None:
        residual_dtype = residual.dtype
    T, D, G = *x.shape, num_groups
    if residual is not None:
        assert residual.shape == (T, D)
    if weight is not None:
        assert weight.shape == (G * D,)
    if bias is not None:
        assert bias.shape == (G * D,)
    # allocate output
    y = torch.empty_like(x, dtype=x.dtype if out_dtype is None else out_dtype)
    if residual is not None or (residual_dtype is not None and residual_dtype != x.dtype):
        res_out = torch.empty(T, D, device=x.device, dtype=residual_dtype)
    else:
        res_out = None
    mean = torch.empty((T,), dtype=torch.float, device=x.device) if not is_rms_norm else None
    rstd = torch.empty((T,), dtype=torch.float, device=x.device)
    # Less than 64KB per feature: enqueue fused kernel
    MAX_FUSED_SIZE = 65536 // x.element_size()
    BD = min(MAX_FUSED_SIZE, triton.next_power_of_2(D))
    if D > BD:
        raise RuntimeError("This layer norm doesn't support feature dim >= 64KB.")
    # heuristics for number of warps

    if D <= 512:
        NB = triton.cdiv(T, 2048)
        def grid(meta): return (triton.cdiv(T, meta['BT']), )
        layer_norm_fwd_kernel[grid](
            x,
            y,
            weight,
            bias,
            residual,
            res_out,
            mean,
            rstd,
            eps,
            T=T,
            G=G,
            D=D,
            BD=BD,
            NB=NB,
            IS_RMS_NORM=is_rms_norm,
            HAS_RESIDUAL=residual is not None,
            STORE_RESIDUAL_OUT=res_out is not None,
            HAS_WEIGHT=weight is not None,
            HAS_BIAS=bias is not None,
        )
    else:
        layer_norm_fwd_kernel1[(T,)](
            x,
            y,
            weight,
            bias,
            residual,
            res_out,
            mean,
            rstd,
            eps,
            G=G,
            D=D,
            BD=BD,
            IS_RMS_NORM=is_rms_norm,
            HAS_RESIDUAL=residual is not None,
            STORE_RESIDUAL_OUT=res_out is not None,
            HAS_WEIGHT=weight is not None,
            HAS_BIAS=bias is not None,
        )
    # res_out is None if residual is None and residual_dtype == input_dtype
    return y, mean, rstd, res_out if res_out is not None else x


def layer_norm_bwd(
    dy: torch.Tensor,
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    mean: torch.Tensor = None,
    rstd: torch.Tensor = None,
    dres: torch.Tensor = None,
    has_residual: bool = False,
    is_rms_norm: bool = False,
    x_dtype: torch.dtype = None,
    recompute_output: bool = False,
    num_groups: int = 1,
):
    T, D, G = *x.shape, num_groups
    assert dy.shape == (T, D)
    if dres is not None:
        assert dres.shape == (T, D)
    if weight is not None:
        assert weight.shape == (G * D,)
    if bias is not None:
        assert bias.shape == (G * D,)
    # allocate output
    dx = torch.empty_like(x) if x_dtype is None else torch.empty(T, D, dtype=x_dtype, device=x.device)
    dres_in = torch.empty_like(x) if has_residual and dx.dtype != x.dtype else None
    y = torch.empty(T, D, dtype=dy.dtype, device=dy.device) if recompute_output else None

    # Less than 64KB per feature: enqueue fused kernel
    MAX_FUSED_SIZE = 65536 // x.element_size()
    BD = min(MAX_FUSED_SIZE, triton.next_power_of_2(D))
    if D > BD:
        raise RuntimeError("This layer norm doesn't support feature dim >= 64KB.")
    # each program handles one group only
    NS = triton.cdiv(get_multiprocessor_count(x.device.index), G) * G
    BS = triton.cdiv(T, NS)
    GS = NS // G

    dw = torch.empty((NS, D), dtype=torch.float, device=weight.device) if weight is not None else None
    db = torch.empty((NS, D), dtype=torch.float, device=bias.device) if bias is not None else None
    grid = (NS,)

    if D <= 512:
        NB = triton.cdiv(T, 2048)
        layer_norm_bwd_kernel[grid](
            x,
            weight,
            bias,
            y,
            dy,
            dx,
            dw,
            db,
            dres,
            dres_in,
            mean,
            rstd,
            T=T,
            G=G,
            D=D,
            BS=BS,
            BD=BD,
            NB=NB,
            GS=GS,
            IS_RMS_NORM=is_rms_norm,
            HAS_DRESIDUAL=dres is not None,
            STORE_DRESIDUAL=dres_in is not None,
            HAS_WEIGHT=weight is not None,
            HAS_BIAS=bias is not None,
        )
    else:
        layer_norm_bwd_kernel1[grid](
            x,
            weight,
            bias,
            y,
            dy,
            dx,
            dw,
            db,
            dres,
            dres_in,
            mean,
            rstd,
            T=T,
            G=G,
            D=D,
            BS=BS,
            BD=BD,
            GS=GS,
            IS_RMS_NORM=is_rms_norm,
            HAS_DRESIDUAL=dres is not None,
            STORE_DRESIDUAL=dres_in is not None,
            HAS_WEIGHT=weight is not None,
            HAS_BIAS=bias is not None,
        )
    dw = dw.view(G, -1, D).sum(1).to(weight).view_as(weight) if weight is not None else None
    db = db.view(G, -1, D).sum(1).to(bias).view_as(bias) if bias is not None else None
    # Don't need to compute dres_in separately in this case
    if has_residual and dx.dtype == x.dtype:
        dres_in = dx
    return (dx, dw, db, dres_in) if not recompute_output else (dx, dw, db, dres_in, y)


class LayerNormFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    def forward(
        ctx,
        x,
        weight,
        bias,
        residual: torch.Tensor = None,
        eps: float = 1e-5,
        prenorm: bool = False,
        residual_in_fp32: bool = False,
        is_rms_norm: bool = False,
        num_groups: int = 1,
    ):
        x_shape_og = x.shape

        if x.shape[-1] % num_groups != 0:
            raise ValueError('num_channels must be divisible by num_groups')
        # reshape input data into 2D tensor
        x = x.reshape(-1, (x.shape[-1] // num_groups))
        if residual is not None:
            assert residual.shape == x_shape_og
            residual = residual.reshape_as(x)
        residual_dtype = (
            residual.dtype
            if residual is not None
            else (torch.float32 if residual_in_fp32 else None)
        )
        y, mean, rstd, res_out = layer_norm_fwd(
            x,
            weight,
            bias,
            eps,
            residual,
            residual_dtype=residual_dtype,
            is_rms_norm=is_rms_norm,
            num_groups=num_groups,
        )
        ctx.save_for_backward(res_out, weight, bias, mean, rstd)
        ctx.x_shape_og = x_shape_og
        ctx.eps = eps
        ctx.is_rms_norm = is_rms_norm
        ctx.num_groups = num_groups
        ctx.has_residual = residual is not None
        ctx.prenorm = prenorm
        ctx.x_dtype = x.dtype
        y = y.reshape(x_shape_og)
        return y if not prenorm else (y, res_out.reshape(x_shape_og))

    @staticmethod
    @input_guard
    def backward(ctx, dy, *args):
        x, weight, bias, mean, rstd = ctx.saved_tensors
        dy = dy.reshape(-1, (dy.shape[-1] // ctx.num_groups))
        assert dy.shape == x.shape
        if ctx.prenorm:
            dresidual = args[0]
            dresidual = dresidual.reshape(-1, x.shape[-1])
            assert dresidual.shape == x.shape
        else:
            dresidual = None
        dx, dw, db, dresidual_in = layer_norm_bwd(
            dy,
            x,
            weight,
            bias,
            mean,
            rstd,
            dresidual,
            ctx.has_residual,
            ctx.is_rms_norm,
            x_dtype=ctx.x_dtype,
            num_groups=ctx.num_groups,
        )
        return (
            dx.reshape(ctx.x_shape_og),
            dw,
            db,
            dresidual_in.reshape(ctx.x_shape_og) if ctx.has_residual else None,
            None,
            None,
            None,
            None,
            None,
        )


def layer_norm(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    residual_in_fp32: bool = False,
    is_rms_norm: bool = False,
):
    return LayerNormFunction.apply(
        x,
        weight,
        bias,
        residual,
        eps,
        prenorm,
        residual_in_fp32,
        is_rms_norm,
    )


def group_norm(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    residual_in_fp32: bool = False,
    is_rms_norm: bool = False,
    num_groups: int = 1,
):
    return LayerNormFunction.apply(
        x,
        weight,
        bias,
        residual,
        eps,
        prenorm,
        residual_in_fp32,
        is_rms_norm,
        num_groups,
    )


def rms_norm(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    residual_in_fp32: bool = False,
):
    return LayerNormFunction.apply(
        x,
        weight,
        bias,
        residual,
        eps,
        prenorm,
        residual_in_fp32,
        True,
    )


def layer_norm_linear(
    x: torch.Tensor,
    norm_weight: torch.Tensor,
    norm_bias: torch.Tensor,
    linear_weight: torch.Tensor,
    linear_bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    residual_in_fp32: bool = False,
    is_rms_norm: bool = False,
    num_groups: int = 1,
):
    return LayerNormLinearFunction.apply(
        x,
        norm_weight,
        norm_bias,
        linear_weight,
        linear_bias,
        residual,
        eps,
        prenorm,
        residual_in_fp32,
        is_rms_norm,
        num_groups,
    )


def rms_norm_linear(
    x: torch.Tensor,
    norm_weight: torch.Tensor,
    norm_bias: torch.Tensor,
    linear_weight: torch.Tensor,
    linear_bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    residual_in_fp32: bool = False,
):
    return layer_norm_linear(
        x=x,
        norm_weight=norm_weight,
        norm_bias=norm_bias,
        linear_weight=linear_weight,
        linear_bias=linear_bias,
        residual=residual,
        eps=eps,
        prenorm=prenorm,
        residual_in_fp32=residual_in_fp32,
        is_rms_norm=True,
    )


def group_norm_linear(
    x: torch.Tensor,
    norm_weight: torch.Tensor,
    norm_bias: torch.Tensor,
    linear_weight: torch.Tensor,
    linear_bias: torch.Tensor,
    residual: torch.Tensor = None,
    eps: float = 1e-5,
    prenorm: bool = False,
    residual_in_fp32: bool = False,
    is_rms_norm: bool = False,
    num_groups: int = 1,
):
    return layer_norm_linear(
        x=x,
        norm_weight=norm_weight,
        norm_bias=norm_bias,
        linear_weight=linear_weight,
        linear_bias=linear_bias,
        residual=residual,
        eps=eps,
        prenorm=prenorm,
        residual_in_fp32=residual_in_fp32,
        is_rms_norm=is_rms_norm,
        num_groups=num_groups,
    )


class LayerNorm(nn.Module):

    def __init__(
        self,
        hidden_size: int,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> LayerNorm:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, residual=None, prenorm=False, residual_in_fp32=False):
        return layer_norm(
            x,
            self.weight,
            self.bias,
            residual=residual,
            eps=self.eps,
            prenorm=prenorm,
            residual_in_fp32=residual_in_fp32,
        )


class GroupNorm(nn.Module):

    def __init__(
        self,
        num_groups: int,
        hidden_size: int,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        is_rms_norm: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> GroupNorm:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        if hidden_size % num_groups != 0:
            raise ValueError('num_channels must be divisible by num_groups')

        self.num_groups = num_groups
        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps
        self.is_rms_norm = is_rms_norm

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.num_groups}, {self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        if self.is_rms_norm:
            s += f", is_rms_norm={self.is_rms_norm}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, residual=None, prenorm=False, residual_in_fp32=False):
        return group_norm(
            x,
            self.weight,
            self.bias,
            residual=residual,
            eps=self.eps,
            prenorm=prenorm,
            residual_in_fp32=residual_in_fp32,
            is_rms_norm=self.is_rms_norm,
            num_groups=self.num_groups,
        )


class RMSNorm(nn.Module):

    def __init__(
        self,
        hidden_size: int,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> RMSNorm:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, residual=None, prenorm=False, residual_in_fp32=False):
        return rms_norm(
            x,
            self.weight,
            self.bias,
            residual=residual,
            eps=self.eps,
            prenorm=prenorm,
            residual_in_fp32=residual_in_fp32,
        )


class LayerNormLinearFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    def forward(
        ctx,
        x,
        norm_weight,
        norm_bias,
        linear_weight,
        linear_bias,
        residual=None,
        eps=1e-5,
        prenorm=False,
        residual_in_fp32=False,
        is_rms_norm=False,
        num_groups=1,
    ):
        x_shape_og = x.shape

        if x.shape[-1] % num_groups != 0:
            raise ValueError('num_channels must be divisible by num_groups')
        # reshape input data into 2D tensor
        x = x.reshape(-1, (x.shape[-1] // num_groups))
        if residual is not None:
            assert residual.shape == x_shape_og
            residual = residual.reshape_as(x)
        residual_dtype = (
            residual.dtype
            if residual is not None
            else (torch.float32 if residual_in_fp32 else None)
        )
        y, mean, rstd, res_out = layer_norm_fwd(
            x,
            norm_weight,
            norm_bias,
            eps,
            residual,
            out_dtype=None if not torch.is_autocast_enabled() else torch.get_autocast_gpu_dtype(),
            residual_dtype=residual_dtype,
            is_rms_norm=is_rms_norm,
            num_groups=num_groups,
        )
        y = y.reshape(x_shape_og)
        dtype = torch.get_autocast_gpu_dtype() if torch.is_autocast_enabled() else y.dtype
        linear_weight = linear_weight.to(dtype)
        linear_bias = linear_bias.to(dtype) if linear_bias is not None else None
        out = F.linear(y.to(linear_weight.dtype), linear_weight, linear_bias)
        # We don't store y, will be recomputed in the backward pass to save memory
        ctx.save_for_backward(res_out, norm_weight, norm_bias, linear_weight, mean, rstd)
        ctx.x_shape_og = x_shape_og
        ctx.eps = eps
        ctx.is_rms_norm = is_rms_norm
        ctx.num_groups = num_groups
        ctx.has_residual = residual is not None
        ctx.prenorm = prenorm
        ctx.x_dtype = x.dtype
        ctx.linear_bias_is_none = linear_bias is None
        return out if not prenorm else (out, res_out.reshape(x_shape_og))

    @staticmethod
    @input_guard
    def backward(ctx, dout, *args):
        x, norm_weight, norm_bias, linear_weight, mean, rstd = ctx.saved_tensors
        dout = dout.reshape(-1, dout.shape[-1])
        dy = F.linear(dout, linear_weight.t())
        dy = dy.reshape(-1, (dy.shape[-1] // ctx.num_groups))
        dlinear_bias = None if ctx.linear_bias_is_none else dout.sum(0)
        assert dy.shape == x.shape
        if ctx.prenorm:
            dresidual = args[0]
            dresidual = dresidual.reshape(-1, x.shape[-1])
            assert dresidual.shape == x.shape
        else:
            dresidual = None
        dx, dnorm_weight, dnorm_bias, dresidual_in, y = layer_norm_bwd(
            dy,
            x,
            norm_weight,
            norm_bias,
            mean,
            rstd,
            dresidual,
            ctx.has_residual,
            ctx.is_rms_norm,
            x_dtype=ctx.x_dtype,
            recompute_output=True,
            num_groups=ctx.num_groups,
        )
        dlinear_weight = torch.einsum("bo,bi->oi", dout, y.view(-1, linear_weight.shape[-1]))
        return (
            dx.reshape(ctx.x_shape_og),
            dnorm_weight,
            dnorm_bias,
            dlinear_weight,
            dlinear_bias,
            dresidual_in.reshape(ctx.x_shape_og) if ctx.has_residual else None,
            None,
            None,
            None,
            None,
            None,
        )


class LayerNormLinear(nn.Module):

    def __init__(
        self,
        hidden_size,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> LayerNormLinear:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, weight, bias, residual=None, prenorm=False, residual_in_fp32=False):
        return layer_norm_linear(
            x=x,
            norm_weight=self.weight,
            norm_bias=self.bias,
            linear_weight=weight,
            linear_bias=bias,
            residual=residual,
            eps=self.eps,
            prenorm=prenorm,
            residual_in_fp32=residual_in_fp32,
            is_rms_norm=False,
        )


class GroupNormLinear(nn.Module):

    def __init__(
        self,
        num_groups: int,
        hidden_size: int,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        is_rms_norm: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> GroupNormLinear:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        if hidden_size % num_groups != 0:
            raise ValueError('num_channels must be divisible by num_groups')

        self.num_groups = num_groups
        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps
        self.is_rms_norm = is_rms_norm

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.num_groups}, {self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        if self.is_rms_norm:
            s += f", is_rms_norm={self.is_rms_norm}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, weight, bias, residual=None, prenorm=False, residual_in_fp32=False):
        return layer_norm_linear(
            x=x,
            norm_weight=self.weight,
            norm_bias=self.bias,
            linear_weight=weight,
            linear_bias=bias,
            residual=residual,
            eps=self.eps,
            prenorm=prenorm,
            residual_in_fp32=residual_in_fp32,
            is_rms_norm=self.is_rms_norm,
            num_groups=self.num_groups,
        )


class RMSNormLinear(nn.Module):

    def __init__(
        self,
        hidden_size,
        elementwise_affine: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> RMSNormLinear:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.hidden_size = hidden_size
        self.elementwise_affine = elementwise_affine
        self.eps = eps

        self.register_parameter("weight", None)
        self.register_parameter("bias", None)
        if elementwise_affine:
            self.weight = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))
            if bias:
                self.bias = nn.Parameter(torch.empty(hidden_size, **factory_kwargs))

        self.reset_parameters()

    def reset_parameters(self):
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}({self.hidden_size}"
        if not self.elementwise_affine:
            s += f", elementwise_affine={self.elementwise_affine}"
        s += f", eps={self.eps}"
        s += ")"
        return s

    def forward(self, x, weight, bias, residual=None, prenorm=False, residual_in_fp32=False):
        return layer_norm_linear(
            x=x,
            norm_weight=self.weight,
            norm_bias=self.bias,
            linear_weight=weight,
            linear_bias=bias,
            residual=residual,
            eps=self.eps,
            prenorm=prenorm,
            residual_in_fp32=residual_in_fp32,
            is_rms_norm=True,
        )


class NormParallel(ParallelStyle):

    def __init__(self, *, sequence_dim: int = 1, use_local_output: bool = False):
        super().__init__()
        self.sequence_sharding = (Shard(sequence_dim),)
        self.use_local_output = use_local_output

    def _replicate_module_fn(
        self, name: str, module: nn.Module, device_mesh: DeviceMesh,
    ):
        for p_name, param in module.named_parameters():
            # simple replication with fixed ones_ init from LayerNorm/RMSNorm, which allow
            # us to simply just use from_local
            replicated_param = torch.nn.Parameter(
                DTensor.from_local(param, device_mesh, [Replicate()], run_check=False),
            )
            module.register_parameter(p_name, replicated_param)

    @staticmethod
    def _prepare_input_fn(sequence_sharding, mod, inputs, device_mesh):
        input_tensor = inputs[0]
        if isinstance(input_tensor, DTensor):
            # if the passed in input DTensor is not sharded on the sequence dim, we need to redistribute it
            if input_tensor.placements != sequence_sharding:
                input_tensor = input_tensor.redistribute(
                    placements=sequence_sharding, async_op=True,
                )
            return input_tensor
        elif isinstance(input_tensor, torch.Tensor):
            # assume the input passed in already sharded on the sequence dim and create the DTensor
            return DTensor.from_local(
                input_tensor, device_mesh, sequence_sharding, run_check=False,
            )
        else:
            raise ValueError(
                f"expecting input of {mod} to be a torch.Tensor or DTensor, but got {input_tensor}",
            )

    @staticmethod
    def _prepare_output_fn(use_local_output, mod, outputs, device_mesh):
        return outputs.to_local() if use_local_output else outputs

    def _apply(self, module: nn.Module, device_mesh: DeviceMesh) -> nn.Module:
        return distribute_module(
            module,
            device_mesh,
            self._replicate_module_fn,
            partial(self._prepare_input_fn, self.sequence_sharding),
            partial(self._prepare_output_fn, self.use_local_output),
        )



"""
Kernels for linear layer with fused activation.
"""


import triton
import triton.language as tl

from .act_kernels import apply_act_func
from .utils import allow_tf32, get_n_stages


def linear_forward_config(
    BLOCK_SIZE_BATCH: int,
    BLOCK_SIZE_IN_FEAT: int,
    BLOCK_SIZE_OUT_FEAT: int,
    GROUP_SIZE_BATCH: int = 8,
    n_warps: int = 4,
    n_stages: int = 2,
    ) -> triton.Config:
    """
    Creates a triton.Config object for linear_forward_kernel
    given meta-parameters for auto-tuning.

    Args:
        BLOCK_SIZE_BATCH: Block size across the batch dimension.
        BLOCK_SIZE_IN_FEAT: Block size across the input feature dimension.
        BLOCK_SIZE_OUT_FEAT: Block size across the output feature dimension.
        GROUP_SIZE_BATCH: Group size across the batch dimension.
        n_warps: Number of warps to use for the kernel when compiled for GPUs.
        n_stages: Number of stages the compiler uses to software-pipeline.
            On GPU architectures older than Ampere, this is fixed at 2.

    Returns:
        Kernel configuration.
    """
    return triton.Config({'BLOCK_SIZE_BATCH': BLOCK_SIZE_BATCH,
                          'BLOCK_SIZE_IN_FEAT': BLOCK_SIZE_IN_FEAT,
                          'BLOCK_SIZE_OUT_FEAT': BLOCK_SIZE_OUT_FEAT,
                          'GROUP_SIZE_BATCH': GROUP_SIZE_BATCH},
                          num_warps=n_warps, num_stages=get_n_stages(n_stages))


@triton.autotune(
    configs=[
        linear_forward_config(32, 32, 32, n_warps=2, n_stages=2),
        linear_forward_config(64, 32, 32, n_warps=2, n_stages=5),
        linear_forward_config(64, 32, 128, n_warps=4, n_stages=4),
        linear_forward_config(64, 32, 256, n_warps=4, n_stages=4),
        linear_forward_config(128, 32, 32, n_warps=4, n_stages=4),
        linear_forward_config(128, 32, 64, n_warps=4, n_stages=4),
        linear_forward_config(128, 32, 128, n_warps=4, n_stages=4),
        linear_forward_config(128, 64, 256, n_warps=8, n_stages=3),
    ],
    key=['batch_dim', 'in_feat_dim', 'out_feat_dim', 'fp16'],
)
@triton.heuristics({'tf32': lambda _: allow_tf32()})
@triton.jit
def linear_forward_kernel(
    input_pointer, weight_pointer, bias_pointer, pre_act_pointer, output_pointer,
    batch_dim, in_feat_dim, out_feat_dim,
    input_batch_stride, input_in_feat_stride,
    weight_in_feat_stride, weight_out_feat_stride,
    pre_act_batch_stride, pre_act_out_feat_stride,
    output_batch_stride, output_out_feat_stride, param,
    add_bias: tl.constexpr, act_func: tl.constexpr, save_pre_act: tl.constexpr,
    fp16: tl.constexpr, tf32: tl.constexpr,
    BLOCK_SIZE_BATCH: tl.constexpr, BLOCK_SIZE_IN_FEAT: tl.constexpr,
    BLOCK_SIZE_OUT_FEAT: tl.constexpr, GROUP_SIZE_BATCH: tl.constexpr,
    ):
    """
    Linearly transforms the input using weights, optionally adding bias
    and fusing an activation function.

    Args:
        input_pointer: Pointer to the input to transform.
            The input must be of shape [batch_dim, in_feat_dim].
        weight_pointer: Pointer to the weights input is transformed by.
            The weights must be of shape [in_feat_dim, out_feat_dim].
        bias_pointer: Pointer to an optional additive bias vector.
            The bias vector, if provided, must be of shape [out_feat_dim].
        pre_act_pointer: Pointer to an optional container the pre-activation input
            is written to if act_func is not None and save_pre_act is True.
            The container, if provided, must be of shape [batch_dim, out_feat_dim].
        output_pointer: Pointer to a container the result is written to.
            The container must be of shape [batch_dim, out_feat_dim].
        batch_dim: Batch dimension of the input and output.
        in_feat_dim: Dimensionality of the input features.
        out_feat_dim: Dimensionality of the output features.
        input_batch_stride: Stride necessary to jump one element along the
            input's batch dimension.
        input_in_feat_stride: Stride necessary to jump one element along the
            input's feature dimension.
        weight_in_feat_stride: Stride necessary to jump one element along the
            weights' input feature dimension.
        weight_out_feat_stride: Stride necessary to jump one element along the
            weights' output feature dimension.
        pre_act_batch_stride: Stride necessary to jump one element along the
            pre-activation input container's batch dimension.
        pre_act_out_feat_stride: Stride necessary to jump one element along the
            pre-activation input container's feature dimension.
        output_batch_stride: Stride necessary to jump one element along the
            output container's batch dimension.
        output_out_feat_stride: Stride necessary to jump one element along the
            output container's feature dimension.
        param: Parameter in the case of parameterized activation functions.
        add_bias: Flag for adding a bias vector.
        act_func: Name of activation function to apply, with None for identity.
            Options are 'sigmoid', 'logsigmoid', 'tanh', 'relu', 'gelu', 'geluapprox', 'silu',
            'softplus', 'softsign', 'tanhshrink', 'leaky_relu', 'elu', 'celu', 'hardshrink',
            and 'softshrink'.
        save_pre_act: Flag for saving the pre-activation input.
        fp16: Flag for loading the input, weights, and bias in FP16.
        tf32: Flag for performing matrix products in TF32.
        BLOCK_SIZE_BATCH: Block size across the batch dimension.
        BLOCK_SIZE_IN_FEAT: Block size across the input feature dimension.
        BLOCK_SIZE_OUT_FEAT: Block size across the output feature dimension.
        GROUP_SIZE_BATCH: Group size across the batch dimension.
    """
    # Programs are blocked together, GROUP_SIZE_BATCH rows at a time,
    # to alleviate L2 miss rates.
    pid = tl.program_id(axis=0)
    n_batch_pids = tl.cdiv(batch_dim, BLOCK_SIZE_BATCH)
    n_out_feat_pids = tl.cdiv(out_feat_dim, BLOCK_SIZE_OUT_FEAT)
    pids_per_group = GROUP_SIZE_BATCH * n_out_feat_pids
    group_id = pid // pids_per_group
    first_batch_pid = group_id * GROUP_SIZE_BATCH
    GROUP_SIZE_BATCH = min(n_batch_pids - first_batch_pid, GROUP_SIZE_BATCH)
    batch_pid = first_batch_pid + (pid % GROUP_SIZE_BATCH)
    out_feat_pid = (pid % pids_per_group) // GROUP_SIZE_BATCH

    batch_offset = (batch_pid * BLOCK_SIZE_BATCH +
                    tl.arange(0, BLOCK_SIZE_BATCH))
    out_feat_offset = (out_feat_pid * BLOCK_SIZE_OUT_FEAT +
                       tl.arange(0, BLOCK_SIZE_OUT_FEAT))

    batch_mask = batch_offset < batch_dim
    out_feat_mask = out_feat_offset < out_feat_dim

    input_pointer += input_batch_stride * batch_offset[:, None]
    weight_pointer += weight_out_feat_stride * out_feat_offset[None, :]

    accum = tl.zeros((BLOCK_SIZE_BATCH, BLOCK_SIZE_OUT_FEAT),
                     dtype=tl.float32)

    for block_ind in range(0, tl.cdiv(in_feat_dim, BLOCK_SIZE_IN_FEAT)):
        in_feat_offset = (block_ind * BLOCK_SIZE_IN_FEAT +
                          tl.arange(0, BLOCK_SIZE_IN_FEAT))
        in_feat_mask = in_feat_offset < in_feat_dim

        curr_input_pointer = (input_pointer +
                              input_in_feat_stride * in_feat_offset[None, :])
        curr_weight_pointer = (weight_pointer +
                               weight_in_feat_stride * in_feat_offset[:, None])

        input_block = tl.load(curr_input_pointer,
                              mask=batch_mask[:, None] & in_feat_mask[None, :])
        weight_block = tl.load(curr_weight_pointer,
                               mask=out_feat_mask[None, :] & in_feat_mask[:, None])

        if fp16:
            input_block = input_block.to(tl.float16)
            weight_block = weight_block.to(tl.float16)

        accum += tl.dot(input_block, weight_block, allow_tf32=tf32)

    if add_bias:
        bias = tl.load(bias_pointer + out_feat_offset,
                       mask=out_feat_mask)

        if fp16:
            bias = bias.to(tl.float16)

        accum += bias[None, :]

    if act_func is not None:
        if save_pre_act:
            pre_act_pointer += (pre_act_batch_stride * batch_offset[:, None] +
                                pre_act_out_feat_stride * out_feat_offset[None, :])
            tl.store(pre_act_pointer, accum,
                     mask=batch_mask[:, None] & out_feat_mask[None, :])

        accum = apply_act_func(accum, None, None, None, param, act_func, False)

    output_pointer += (output_batch_stride * batch_offset[:, None] +
                       output_out_feat_stride * out_feat_offset[None, :])
    tl.store(output_pointer, accum,
             mask=batch_mask[:, None] & out_feat_mask[None, :])



import importlib
import logging
import math
import os
import textwrap
from typing import Callable, List, Mapping, Tuple, Union

import torch

from flag_gems.utils.code_cache import cache_dir
from flag_gems.utils.code_utils import IndentedBuffer, write_atomic

from ..utils import TOTAL_CORE_NUM
from .vstack import vstack

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


def get_dtype_size(dtype):
    try:
        return torch.finfo(dtype).bits // 8
    except TypeError:
        try:
            return torch.iinfo(dtype).bits // 8
        except TypeError:
            if dtype == torch.bool:
                return 1
            else:
                raise ValueError(f"Unsupported dtype: {dtype}")


class StackKernelCode(IndentedBuffer):
    """
    Stack kernel template.
    """

    overloads: Mapping[str, Callable] = {}

    def __init__(self):
        self.pid = os.getpid()
        self.cache = self.overloads
        self.kernel_name = "_stack_jit_kernel"
        self.wrapper_func_name = "_wrapper"
        super(StackKernelCode, self).__init__()

    def __imports(self):
        """Generate imports for the kernel code."""
        tpl = """\
            import math
            import torch
            import triton
            from triton import language as tl
            from typing import List, Tuple, Union
            from flag_gems.utils import libentry
            from flag_gems.runtime.backend import vendor_module
            TOTAL_CORE_NUM = vendor_module.TOTAL_CORE_NUM
            MAX_NRAM_SIZE = vendor_module.MAX_NRAM_SIZE

            """
        self.tpl(textwrap.dedent(tpl))

    def __wrapper(self):
        """Generate wrapper function for the kernel code."""
        self.newline()
        tpl = """\
            def {wrapper_name}(
                tensors: Union[Tuple[torch.Tensor, ...], List[torch.Tensor]], dim: int = 0
            ) -> torch.Tensor:
                if len(tensors) == 0:
                    raise RuntimeError("stack expected a non-empty TensorList")

                inp_shapes = [list(_.shape) for _ in tensors]
                inp0_shape = inp_shapes[0]
                for i, s in enumerate(inp_shapes[1:]):
                    if (dim < -tensors[i + 1].dim() - 1) or (dim > tensors[i + 1].dim()):
                        raise IndexError(
                            "Dimension out of range (expected to be in range of [{{}}, {{}}], but got {{}})".format(
                                -tensors[i + 1].dim() - 1, tensors[i + 1].dim(), dim
                            )
                        )
                    if s != inp0_shape:
                        raise RuntimeError(
                            f"stack expects each tensor to be equal size, \
                                but got {{inp0_shape}} at entry 0 and {{s}} at entry {{i+1}}"
                        )

                if dim < 0:
                    dim = dim + len(inp0_shape) + 1
                out_shape = inp0_shape[:dim] + [len(tensors)] + inp0_shape[dim:]
                high = int(math.prod(out_shape[:dim]))
                low = int(math.prod(out_shape[dim+1:]))
                tensor_num = len(tensors)
                out0 = torch.empty(out_shape, dtype=tensors[0].dtype, device=tensors[0].device)
                def grid(meta):
                    if meta['BLOCK_SIZE']>0:
                        task_x = high
                        task_y = tensor_num
                        task_z = triton.cdiv(low ,meta['BLOCK_SIZE'])
                        return (task_x, task_y, task_z)
                    else:
                        total_task = high * tensor_num
                        if meta['LOW_NUM']>0:
                            core_used = triton.cdiv(total_task // meta['LOW_NUM'], meta['TASK_PER_CORE'])
                        elif meta['N_LOW_NUM']>0:
                            core_used = triton.cdiv(high, meta['TASK_PER_CORE'])
                        return (core_used,)
                {kernel_name}[grid](
                    out0,
                    *tensors,
                    high,
                    tensor_num,
                    low,
                    )
                return out0
            """
        self.tpl(
            textwrap.dedent(tpl),
            wrapper_name=self.wrapper_func_name,
            kernel_name=self.kernel_name,
        )

    def __config(self, tensor_num, high, low, dtype):
        """Generate config for the kernel code."""
        dtyp_bytest = get_dtype_size(dtype)
        # Since the kernel has three branches, each branch has its own parameters,
        # so for a certain branch, the other parameters can be directly set to zero.

        # 1）N_LOW_NUM branch： NRAM can hold at least one set of `tensor_num * low * dtyp_bytest`.
        #    This parameter is used to indicate how many `tensor_num * low` are processed in a single core.

        # 2) LOW_NUM branch: NRAM can hold at least one set of `low * dtyp_bytest`,
        #    but cannot hold tensor_num * low * dtyp_bytest.
        #    This parameter is used to indicate how many `low` are processed in a single core.

        # 3) BLOCK_SIZE branch: NRAM is not enough to store a set of `low`,
        #    so it can only loop multiple times to process a set of `low`.
        #    This parameter is used to indicates how many elements to load
        #    at a time when looping over and processing low.
        tpl = """\
            def cfggen():
                N_LOW_NUM = {n_low_num_options}
                LOW_NUM = {low_num_options}
                BLOCK_SIZE = {block_size_options}
                warps = [1]
                num_stages = {num_stages}
                configs = [
                    triton.Config(
                        {{
                            "BLOCK_SIZE": block_size,
                            "N_LOW_NUM": n_low_num,
                            "LOW_NUM": low_num,
                        }},
                        num_warps=w,
                        num_stages=s)
                    for block_size in BLOCK_SIZE
                    for n_low_num in N_LOW_NUM
                    for low_num in LOW_NUM
                    for w in warps for s in num_stages
                ]
                return configs
            performance_related_keys = {keys}
            """

        # If `tensor_num * low * dtyp_bytest` is less than `nram_threshold`,
        # use N_LOW_NUM branch, otherwise LOW_NUM branch.
        nram_threshold = 170000
        # The maximum number of elements in triton is 1048576.
        max_elements_num = 1048576
        # after removing the overhead of pipeline and temporary variables.
        if tensor_num * low * dtyp_bytest <= nram_threshold:
            n_low_per_core = math.ceil(high / TOTAL_CORE_NUM)
            limited_by_nram = nram_threshold // dtyp_bytest // (tensor_num * low)
            limited_by_triton = max_elements_num // (tensor_num * low)
            best_opt = min(n_low_per_core, limited_by_triton, limited_by_nram)
            self.tpl(
                textwrap.dedent(tpl),
                n_low_num_options=f"{[best_opt]}",
                low_num_options=r"[0]",
                block_size_options=r"[0]",
                num_stages=r"[1]",
                keys=r'["high"]',
            )
        elif low * dtyp_bytest <= nram_threshold:
            self.tpl(
                textwrap.dedent(tpl),
                n_low_num_options=r"[0]",
                low_num_options=r"[1,2,3]",
                block_size_options=r"[0]",
                num_stages=f"{[1]}",
                keys=r'["high", "tensor_num", "low"]',
            )
        else:
            self.tpl(
                textwrap.dedent(tpl),
                n_low_num_options=r"[0]",
                low_num_options=r"[0]",
                block_size_options=r"[8192, 16384, 32768, 65536, 131072, 262144]",
                num_stages=r"[1]",
                keys=r'["low"]',
            )

    def __kernel(self, tensor_num):
        """Generate kernel code body."""
        tpl = """\
            def stack_heuristics(args, need_key):
                ret = {{
                    'TASK_PER_CORE': 0,
                    'TASK_LAST_CORE_REPEAT': 0,
                    'TASK_LAST_CORE_REMAIN': 0,
                }}
                total_task = args['high']*args['tensor_num']
                if args['LOW_NUM']>0:
                    LOW_NUM = args['LOW_NUM'] if total_task > args['LOW_NUM'] else total_task
                    ret['TASK_PER_CORE'] = triton.cdiv(total_task // LOW_NUM, TOTAL_CORE_NUM)
                    assert ret['TASK_PER_CORE']>0, ret['TASK_PER_CORE']
                    core_used = triton.cdiv(total_task // LOW_NUM, ret['TASK_PER_CORE'])
                    task_last_core = total_task-(core_used-1)*ret['TASK_PER_CORE']*LOW_NUM
                    ret['TASK_LAST_CORE_REPEAT'] = task_last_core//LOW_NUM
                    ret['TASK_LAST_CORE_REMAIN'] = task_last_core%LOW_NUM
                elif args['N_LOW_NUM']>0:
                    ret['TASK_PER_CORE'] = triton.cdiv(args['high'], TOTAL_CORE_NUM)
                    core_used = triton.cdiv(args['high'], ret['TASK_PER_CORE'])
                    ret['TASK_LAST_CORE_REPEAT'] = args['high'] -(core_used-1)*ret['TASK_PER_CORE']
                return ret[need_key]

            @triton.jit()
            def load_trans_store(
                low: tl.constexpr,
                tensor_num: tl.constexpr,
                {tensors},
                offset,
                buffer,
                buffer_offset,
                output,
                out_offset,
            ):
                if low >64:
            {low_gt_64_code}
                    tl.store(output+offset*tensor_num+out_offset, buffer)
                else:
            {low_le_64_code}
                    tl.store(output+offset*tensor_num+out_offset, tl.trans(buffer, 1, 0, 2))

            @triton.jit()
            def load_and_store(
                output_ptr,
                buffer,
                buffer_offset,
                task_id,
                LOW_NUM: tl.constexpr,
                low: tl.constexpr,
                LOW_OFFSET: tl.constexpr,
                tensor_num: tl.constexpr,
                {tensors}
            ):
                for low_idx in tl.range(LOW_NUM):
                    cur_low_id = task_id + low_idx
                    tensor_idx = cur_low_id%tensor_num
                    high_idx = cur_low_id//tensor_num
                    load_start = high_idx *low
            {load_and_store_code}
                tl.store(output_ptr+buffer_offset, buffer)

            @libentry()
            @triton.autotune(configs=cfggen(), key=performance_related_keys)
            @triton.heuristics(
                {{
                "TASK_PER_CORE": lambda args: stack_heuristics(args, "TASK_PER_CORE"),
                "TASK_LAST_CORE_REPEAT": lambda args: stack_heuristics(args, "TASK_LAST_CORE_REPEAT"),
                "TASK_LAST_CORE_REMAIN": lambda args: stack_heuristics(args, "TASK_LAST_CORE_REMAIN"),
                }}
            )
            @triton.jit()
            def {kernel_name}(
                output,
                {tensors},
                high: tl.constexpr,
                tensor_num: tl.constexpr,
                low: tl.constexpr,
                N_LOW_NUM: tl.constexpr,
                LOW_NUM: tl.constexpr,
                TASK_PER_CORE: tl.constexpr,
                TASK_LAST_CORE_REPEAT: tl.constexpr,
                TASK_LAST_CORE_REMAIN: tl.constexpr,
                BLOCK_SIZE: tl.constexpr):
                if N_LOW_NUM>0:
                    # The memory space is sufficient to hold at least one set of "tensor_num* low * type_bytes"
                    core_idx = tl.program_id(0)
                    core_used = tl.num_programs(0)
                    if core_idx>=core_used:
                        return
                    in_offset = core_idx*TASK_PER_CORE*low
                    if low >64:
                        buffer_repeat = tl.empty(shape=[N_LOW_NUM, tensor_num, low], dtype=output.dtype.element_ty)
                    else:
                        buffer_repeat = tl.empty(shape=[tensor_num, N_LOW_NUM, low], dtype=output.dtype.element_ty)
                    buffer_repeat_offset = tl.arange(0, N_LOW_NUM)[:, None]*low+tl.arange(0, low)[None,:]
                    out_repeat_offset= \\
                        tl.arange(0, N_LOW_NUM)[:,None,None]*low*tensor_num+\\
                        tl.arange(0, tensor_num)[None,:,None]*low+\\
                        tl.arange(0, low)[None, None,:]
                    if core_idx !=core_used -1:
                        for repeat_idx in range(TASK_PER_CORE//N_LOW_NUM):
                            repeat_offset = in_offset + repeat_idx*N_LOW_NUM*low
                            load_trans_store(low, tensor_num, {tensors},repeat_offset, buffer_repeat,\\
                                buffer_repeat_offset,output,out_repeat_offset)
                        if (TASK_PER_CORE%N_LOW_NUM) > 0:
                            normal_remain_offset = in_offset + (TASK_PER_CORE//N_LOW_NUM)*N_LOW_NUM*low
                            if low >64:
                                buffer_normal_remain = tl.empty(shape=[TASK_PER_CORE%N_LOW_NUM,tensor_num, low], \\
                                    dtype=output.dtype.element_ty)
                            else:
                                buffer_normal_remain = tl.empty(shape=[tensor_num,TASK_PER_CORE%N_LOW_NUM, low], \\
                                    dtype=output.dtype.element_ty)
                            buffer_normal_remain_offset = tl.arange(0, TASK_PER_CORE%N_LOW_NUM)[:, None]*low + \\
                                tl.arange(0, low)[None,:]
                            out_normal_remain_offset= \\
                                tl.arange(0, TASK_PER_CORE%N_LOW_NUM)[:,None,None]*low*tensor_num+\\
                                tl.arange(0, tensor_num)[None,:,None]*low+\\
                                tl.arange(0, low)[None, None,:]
                            load_trans_store(low, tensor_num, {tensors},normal_remain_offset, buffer_normal_remain,\\
                                buffer_normal_remain_offset,output,out_normal_remain_offset)
                    else:
                        for repeat_idx in range(TASK_LAST_CORE_REPEAT//N_LOW_NUM):
                            repeat_offset = in_offset + repeat_idx*N_LOW_NUM*low
                            load_trans_store(low, tensor_num, {tensors},repeat_offset, buffer_repeat,\\
                                buffer_repeat_offset,output,out_repeat_offset)
                        if (TASK_LAST_CORE_REPEAT%N_LOW_NUM) >0 :
                            last_core_remain_offset = in_offset + (TASK_LAST_CORE_REPEAT//N_LOW_NUM)*N_LOW_NUM*low
                            if low >64:
                                buffer_last_core_remain = \\
                                    tl.empty(shape=[TASK_LAST_CORE_REPEAT%N_LOW_NUM,tensor_num, low], \\
                                        dtype=output.dtype.element_ty)
                            else:
                                buffer_last_core_remain = \\
                                    tl.empty(shape=[tensor_num,TASK_LAST_CORE_REPEAT%N_LOW_NUM, low], \\
                                    dtype=output.dtype.element_ty)
                            buffer_last_core_remain_offset = \\
                                tl.arange(0, TASK_LAST_CORE_REPEAT%N_LOW_NUM)[:, None]*low + \\
                                tl.arange(0, low)[None,:]
                            out_last_core_remain_offset= \\
                                tl.arange(0, TASK_LAST_CORE_REPEAT%N_LOW_NUM)[:,None,None]*low*tensor_num+\\
                                tl.arange(0, tensor_num)[None,:,None]*low+\\
                                tl.arange(0, low)[None, None,:]
                            load_trans_store(low, tensor_num, {tensors},last_core_remain_offset, \\
                                buffer_last_core_remain, buffer_last_core_remain_offset, \\
                                output,out_last_core_remain_offset)
                elif LOW_NUM>0:
                    # The memory space is sufficient to hold at least one set of "low * type_bytes"
                    core_idx = tl.program_id(0)
                    core_used = tl.num_programs(0)
                    if core_idx>=core_used:
                        return
                    dtype = output.dtype.element_ty
                    buffer = tl.empty(shape=[LOW_NUM,low], dtype=dtype)
                    buffer_offset = tl.arange(0, LOW_NUM)[:,None]*low+tl.arange(0, low)[None,:]
                    LOW_OFFSET = tl.arange(0, low)
                    if core_idx != core_used-1:
                        for cycles_idx in range(TASK_PER_CORE):
                            task_id = core_idx*TASK_PER_CORE*LOW_NUM+cycles_idx*LOW_NUM
                            out_ptr = output + task_id*low
                            load_and_store(
                                out_ptr,
                                buffer,
                                buffer_offset,
                                task_id,
                                LOW_NUM,
                                low,
                                LOW_OFFSET,
                                tensor_num,
                                {tensors}
                            )
                    else:
                        base_task_id = core_idx*TASK_PER_CORE*LOW_NUM
                        for cycles_idx in range(TASK_LAST_CORE_REPEAT):
                            task_id= base_task_id+cycles_idx*LOW_NUM
                            out_ptr = output + task_id*low
                            load_and_store(
                                out_ptr,
                                buffer,
                                buffer_offset,
                                task_id,
                                LOW_NUM,
                                low,
                                LOW_OFFSET,
                                tensor_num,
                                {tensors}
                            )
                        task_id = base_task_id+TASK_LAST_CORE_REPEAT*LOW_NUM
                        output_ptr = output + task_id*low
                        for low_idx in tl.range(TASK_LAST_CORE_REMAIN):
                            cur_low_id = task_id + low_idx
                            tensor_idx = cur_low_id%tensor_num
                            high_idx = cur_low_id//tensor_num
                            load_start = high_idx *low
            {low_num_gt_0_last_core_code}
                        tl.store(output_ptr+buffer_offset, buffer, mask=buffer_offset<TASK_LAST_CORE_REMAIN*low)
                elif BLOCK_SIZE>0:
                    # Insufficient memory space to hold a set of "low* type_bytes"
                    high_idx = tl.program_id(0)
                    tensor_idx = tl.program_id(1)
                    output_ptr = output + high_idx*(low*tensor_num)+tensor_idx*low
                    offset_in_loop = tl.program_id(2)*BLOCK_SIZE+tl.arange(0, BLOCK_SIZE)
                    x = tl.empty(shape=[BLOCK_SIZE,],dtype=output.dtype.element_ty)
            {block_size_gt_0_code}
                    tl.store(output_ptr+offset_in_loop, x, mask=offset_in_loop<low)
            """

        def add_indent(cleaned_str, indent_size):
            return "\n".join(
                [f"{' ' * indent_size}{line}" for line in cleaned_str.split("\n")]
            )

        tensors = ", ".join([f"in_{idx}" for idx in range(tensor_num)])
        load_form_inputs = textwrap.dedent(
            """\
            if tensor_idx == 0:
                buffer[low_idx,:] = tl.load(in_0+load_start+LOW_OFFSET)\n"""
            + "\n".join(
                [
                    f"""\
            elif tensor_idx == {idx}:
                buffer[low_idx,:] = tl.load(in_{idx}+load_start+LOW_OFFSET)"""
                    for idx in range(1, tensor_num - 1)
                ]
            )
            + "\n"
            + f"""\
            else:
                buffer[low_idx,:] = tl.load(in_{tensor_num - 1}+load_start+LOW_OFFSET)"""
        )
        self.tpl(
            textwrap.dedent(tpl),
            kernel_name=self.kernel_name,
            tensors=tensors,
            low_gt_64_code="\n".join(
                [
                    f"{' ' * 8}buffer[:,{idx},:]=tl.load(in_{idx}+offset+buffer_offset)"
                    for idx in range(tensor_num)
                ]
            ),
            low_le_64_code="\n".join(
                [
                    f"{' ' * 8}buffer[{idx},:,:]=tl.load(in_{idx}+offset+buffer_offset)"
                    for idx in range(tensor_num)
                ]
            ),
            load_and_store_code=add_indent(load_form_inputs, 8),
            low_num_gt_0_last_core_code=add_indent(load_form_inputs, 16),
            block_size_gt_0_code=add_indent(
                textwrap.dedent(
                    """\
                if tensor_idx == 0:
                    x = tl.load(in_0+high_idx *low+offset_in_loop,mask=offset_in_loop<low)\n"""
                    + "\n".join(
                        [
                            f"""\
                elif tensor_idx == {idx}:
                    x = tl.load(in_{idx}+high_idx *low+offset_in_loop,mask=offset_in_loop<low)"""
                            for idx in range(1, tensor_num - 1)
                        ]
                    )
                    + "\n"
                    + f"""\
                else:
                    x = tl.load(in_{tensor_num - 1}+high_idx *low+offset_in_loop,mask=offset_in_loop<low)"""
                ),
                8,
            ),
        )

    def __gen_code(self, tensor_num, high, low, dtype):
        """Entry point for code generation of stack."""
        # generate imports.
        self.__imports()
        # generate config.
        self.__config(tensor_num, high, low, dtype)
        # generate kernel.
        self.__kernel(tensor_num)
        # generate wrapper function.
        self.__wrapper()

    def __call__(
        self, tensors: Union[Tuple[torch.Tensor, ...], List[torch.Tensor]], dim: int = 0
    ) -> torch.Tensor:
        assert dim != 0, "StackKernel template does not optimize `dim=0`."
        tensor_num = len(tensors)
        inp0_shape = list(tensors[0].shape)
        out_shape = inp0_shape[:dim] + [len(tensors)] + inp0_shape[dim:]
        high = int(math.prod(out_shape[:dim]))
        low = int(math.prod(out_shape[dim + 1 :]))
        dtype = tensors[0].dtype
        self.kernel_name = f"{self.kernel_name}_num_{tensor_num}"
        key = f"num_{tensor_num}_high_{high}_low_{low}_dtype_{dtype}"
        for tensor in tensors[1:]:
            assert tensor.dtype == dtype, f"{tensor.dtype} != {dtype}"
        if key not in self.cache:
            # generate code and cache.
            self.__gen_code(tensor_num, high, low, dtype)
            file_name = f"{cache_dir()}/stack_{key}_pid_{self.pid}.py"
            write_atomic(file_name, self.getvalue())
            # load
            spec = importlib.util.spec_from_file_location(
                f"_gen_module_{key}_pid_{self.pid}", file_name
            )
            m = importlib.util.module_from_spec(spec)
            # do not expose it to sys.modules
            # sys.modules["_add_module"] = m
            spec.loader.exec_module(m)
            overload = getattr(m, self.wrapper_func_name)
            self.cache[key] = overload
        overload = self.cache[key]
        return overload(tensors, dim)


def stack(
    tensors: Union[Tuple[torch.Tensor, ...], List[torch.Tensor]], dim: int = 0
) -> torch.Tensor:
    logger.debug("GEMS_CAMBRICON STACK")

    if len(tensors) == 0:
        raise RuntimeError("stack expected a non-empty TensorList")

    inp_shapes = [list(_.shape) for _ in tensors]
    inp0_shape = inp_shapes[0]
    for i, s in enumerate(inp_shapes[1:]):
        if (dim < -tensors[i + 1].dim() - 1) or (dim > tensors[i + 1].dim()):
            raise IndexError(
                "Dimension out of range (expected to be in range of [{}, {}], but got {})".format(
                    -tensors[i + 1].dim() - 1, tensors[i + 1].dim(), dim
                )
            )
        if s != inp0_shape:
            raise RuntimeError(
                f"stack expects each tensor to be equal size, but got {inp0_shape} at entry 0 and {s} at entry {i + 1}"
            )

    if dim < 0:
        dim = dim + len(inp0_shape) + 1
    out_shape = inp0_shape[:dim] + [len(tensors)] + inp0_shape[dim:]
    if dim == 0:
        return vstack(tensors).view(out_shape)
    tensors = [
        tensor if tensor.is_contiguous() else tensor.contiguous() for tensor in tensors
    ]
    return StackKernelCode()(tensors, dim)



"""
Input Resolver

ZERO HARDCODE: All bindings come from topology.json connections.
ZERO SEMANTIC: No knowledge of "text_encoder", "vae" - only connection resolution.

Resolves component inputs by following topology connections from global
variables and component outputs.
"""

from typing import Any, Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    from neurobrix.core.runtime.resolution.variable_resolver import VariableResolver


class InputResolver:
    """
    Resolves component inputs from topology connections.

    This class is responsible for:
    1. Looking up connections from the pre-indexed connections map
    2. Resolving source chains (trying sources in order until one resolves)
    3. Resolving individual sources (global variables or component outputs)

    ZERO HARDCODE: All bindings from topology.json connections.
    ZERO SEMANTIC: No domain knowledge - just follows connection graph.
    """

    def __init__(
        self,
        variable_resolver: 'VariableResolver',
        connections_index: Dict[str, Dict[str, List[str]]],
        topology: Dict[str, Any],
        loop_id: str
    ):
        """
        Initialize input resolver.

        Args:
            variable_resolver: VariableResolver instance for accessing resolved values
            connections_index: Pre-indexed connections {comp: {input: [sources]}}
            topology: Full topology dict for loop definitions
            loop_id: Current loop identifier for timestep resolution
        """
        self._variable_resolver = variable_resolver
        self._connections_index = connections_index
        self._topology = topology
        self._loop_id = loop_id

    def resolve_component_inputs(self, comp_name: str) -> Dict[str, Any]:
        """
        Resolve component inputs from topology connections.

        ZERO HARDCODE: All bindings from topology.json connections.
        For iterative patterns, tries sources in order (global first, then feedback).

        UNIFIED RESOLUTION: Same pattern for all flow types (diffusers and forward_pass).
        Connections are enriched at import time from graph inputs, so runtime
        uses the same resolve_source_chain() for all models.

        FALLBACK RESOLUTION: If no connections declared but component interface
        specifies inputs, try to resolve from variable_resolver.resolved.

        Args:
            comp_name: Name of component to resolve inputs for

        Returns:
            Dict mapping input names to resolved values
        """
        # Check if component has declared connections
        # UNIVERSAL: Use connections from topology (same pattern for all model types)
        # Connections are generated at trace time (Phase C) for all models including forward_pass
        if comp_name in self._connections_index:
            connections = self._connections_index[comp_name]
            if connections:  # Has actual connection mappings
                resolved = {}
                for input_name, sources in connections.items():
                    # sources is a list - try each in order until one resolves
                    resolved[input_name] = self.resolve_source_chain(sources)
                return resolved

        # FALLBACK: If no connections, try to resolve from variable_resolver.resolved
        # This handles cases where tokenization stores input_ids/attention_mask
        comp_info = self._topology.get("components", {}).get(comp_name, {})
        interface_inputs = comp_info.get("interface", {}).get("inputs", [])

        if interface_inputs:
            resolved = {}
            for input_name in interface_inputs:
                # Try multiple resolution paths
                for key in [input_name, f"global.{input_name}"]:
                    if key in self._variable_resolver.resolved:
                        resolved[input_name] = self._variable_resolver.resolved[key]
                        break
            if resolved:
                return resolved

        # Component has no declared inputs - may use global defaults
        return {}

    def resolve_source_chain(self, sources: List[str]) -> Any:
        """
        Try to resolve sources in order. First available wins.

        This enables iterative patterns:
        global.latents (initial) → transformer.output_0 (feedback)

        Args:
            sources: List of source identifiers to try in order

        Returns:
            First successfully resolved value

        Raises:
            RuntimeError: If no source could be resolved (ZERO FALLBACK)
        """
        last_error = None
        for source in sources:
            try:
                return self.resolve_source(source)
            except (RuntimeError, KeyError) as e:
                last_error = e
                continue

        # None resolved
        raise RuntimeError(
            f"ZERO FALLBACK: None of the sources resolved: {sources}\n"
            f"Last error: {last_error}"
        )

    def resolve_source(self, source: str) -> Any:
        """
        Resolve a source port to its value.

        Sources can be:
        - "global.xxx" - global variable
        - "component.output" - output from another component

        VARIABLE EMBARQUÉE: index_variable from topology.flow.loop, not hardcoded.

        Args:
            source: Source identifier string

        Returns:
            Resolved value from variable resolver

        Raises:
            RuntimeError: If source cannot be resolved (ZERO FALLBACK)
        """
        if source.startswith("global."):
            # VARIABLE EMBARQUÉE: Check if this is the loop index variable (e.g., timestep)
            loop_def = self._topology.get("flow", {}).get("loop", {})
            index_variable = loop_def.get("index_variable")

            if index_variable and source == index_variable:
                return self._variable_resolver.loop_state.get(self._loop_id)
            return self._variable_resolver.get(source)

        # Component output
        if source in self._variable_resolver.resolved:
            return self._variable_resolver.resolved[source]

        raise RuntimeError(
            f"ZERO FALLBACK: Source '{source}' not found in resolved variables.\n"
            f"Available: {list(self._variable_resolver.resolved.keys())}"
        )

    def get_connections(self, comp_name: str) -> Dict[str, List[str]]:
        """
        Get raw connection mapping for a component.

        Args:
            comp_name: Component name

        Returns:
            Dict mapping input names to source lists
        """
        return self._connections_index.get(comp_name, {})

    def update_loop_id(self, loop_id: str) -> None:
        """
        Update the current loop identifier.

        Used when switching between loops or entering iteration.

        Args:
            loop_id: New loop identifier
        """
        self._loop_id = loop_id



import torch
import triton
import triton.language as tl
from neurobrix.kernels.registry import register_kernel


@triton.jit
def _tanh_kernel(
    x_ptr, out_ptr,
    n_elements,
    BLOCK_SIZE: tl.constexpr
):
    """Numerically stable tanh implementation."""
    pid = tl.program_id(0)
    offsets = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements

    x = tl.load(x_ptr + offsets, mask=mask)

    # Numerically stable tanh:
    # For x >= 0: tanh(x) = 1 - 2/(exp(2x) + 1)
    # For x < 0: tanh(x) = 2/(exp(-2x) + 1) - 1
    # This avoids overflow in exp(2x) for large positive x

    # Clamp to avoid extreme values
    x_clamped = tl.minimum(tl.maximum(x, -20.0), 20.0)

    # Use the standard formula with clamped input
    exp_neg_2x = tl.exp(-2.0 * tl.abs(x_clamped))
    tanh_abs = (1.0 - exp_neg_2x) / (1.0 + exp_neg_2x)

    # Apply sign
    out = tl.where(x >= 0, tanh_abs, -tanh_abs)

    tl.store(out_ptr + offsets, out, mask=mask)


@register_kernel(family="audio", vendor="nvidia", tier="common", op_name="tanh")
def tanh_kernel(x: torch.Tensor) -> torch.Tensor:
    """Tanh activation - pure Triton."""
    assert x.is_cuda
    x = x.contiguous()
    output = torch.empty_like(x)
    n = x.numel()
    if n == 0:
        return output

    BLOCK = 1024
    grid = (triton.cdiv(n, BLOCK),)

    with torch.cuda.device(x.device):
        _tanh_kernel[grid](x, output, n, BLOCK_SIZE=BLOCK, num_warps=4)

    return output




import logging
from typing import List

import torch
import triton
import triton.language as tl

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


@triton.jit
def index_kernel_func(
    input_ptr,
    stride: tl.constexpr,
    index_len,
    index_ptr,
    out_ptr,
    BLOCK_SIZE: tl.constexpr,
    MAX_DATA_SIZE: tl.constexpr,
):
    pid0 = tl.program_id(axis=0)

    for i in range(0, BLOCK_SIZE):
        offset = pid0 * BLOCK_SIZE + i

        if offset < index_len:
            in_start_index = tl.load(index_ptr + offset) * stride
            out_start_offset = offset * stride
            loop_num = (stride - 1) // MAX_DATA_SIZE + 1

            for loop_idx in range(0, loop_num):
                inner_offset = loop_idx * MAX_DATA_SIZE + tl.arange(0, MAX_DATA_SIZE)
                mask = inner_offset < stride
                cur_value = tl.load(
                    input_ptr + in_start_index + inner_offset, mask=mask
                )
                tl.store(
                    out_ptr + out_start_offset + inner_offset, cur_value, mask=mask
                )


def index_wrapper(input, indices, out):
    input_shape = input.shape
    input_dim = len(input_shape)
    indices_dim = len(indices)

    stride = 1
    for i in range(0, input_dim - indices_dim):
        stride *= input_shape[input_dim - i - 1]

    index_len = indices[0].numel()

    actual_index = indices[0]
    for idx in range(0, indices_dim - 1):
        actual_index = actual_index * input_shape[idx + 1] + indices[idx + 1]

    BLOCK_SIZE = 32
    MAX_DATA_SIZE = 16 * 1024

    grid = lambda meta: (triton.cdiv(index_len, meta["BLOCK_SIZE"]),)

    index_kernel_func[grid](
        input,
        stride,
        index_len,
        actual_index,
        out,
        BLOCK_SIZE=BLOCK_SIZE,
        MAX_DATA_SIZE=MAX_DATA_SIZE,
    )


def get_max_rank_shape(indices: List[torch.Tensor]) -> List[int]:
    max_rank = max([len(index.shape) for index in indices])
    shape = [0 for _ in range(max_rank)]
    for i in range(max_rank):
        max_num = 0
        for index in indices:
            axis = len(index.shape) - 1 - i
            if axis >= 0:
                max_num = max(max_num, index.shape[axis])  #
        shape[max_rank - 1 - i] = max_num
    return shape


def broadcast_indices(indices, target_shape):
    for i, index in enumerate(indices):
        if tuple(index.shape) != tuple(target_shape):
            indices[i] = torch.broadcast_to(index, target_shape)


def index(inp, indices):
    logger.debug("GEMS_ASCEND INDEX")
    indices = list(indices)

    target_shape = get_max_rank_shape(indices)
    broadcast_indices(indices, target_shape)
    target_shape += inp.shape[len(indices) :]
    out = torch.empty(target_shape, dtype=inp.dtype, device=inp.device)

    index_wrapper(inp, indices, out)
    return out



"""
Fine-Grained Pipeline (FGP) Execution Strategy

Block-level distribution across multiple GPUs with NVLink transfers.

Key differences from Pipeline:
- Pipeline: Component-level distribution (text_encoder → GPU0, transformer → GPU1)
- FGP: Block-level distribution within a single component (blocks 0-7 → GPU0, 8-15 → GPU1)

FGP is needed when a single component is too large for any single GPU.
The component's blocks (layers) are distributed across GPUs, with tensor
transfers between blocks at GPU boundaries.

ARCHITECTURE:
- Prism provides shard_map: {"block.0.*": "cuda:0", "block.1.*": "cuda:1", ...}
- WeightLoader loads weights directly to their assigned GPUs
- GraphExecutor runs in FGP mode with device alignment per op
- This strategy handles:
  * Setting FGP mode on executors
  * Coordinating weight loading with shard maps
  * Managing tensor transfers between GPUs during execution
"""

from typing import Dict, Any, Optional, List, Set, Tuple
import torch

from .base import ExecutionStrategy, StrategyContext


class FGPStrategy(ExecutionStrategy):
    """
    Fine-Grained Pipeline execution across multiple GPUs.

    Blocks within a component are distributed across GPUs.
    Tensor transfers happen at block boundaries when crossing GPU boundaries.

    Unlike Pipeline (component-level), FGP operates at block-level within
    a single component, enabling models where even one component exceeds
    single GPU memory.
    """

    def __init__(self, context: StrategyContext, strategy_name: str = "fgp_nvlink"):
        super().__init__(context, strategy_name)

        # Track devices and their assignments
        self.devices: List[str] = context.get_all_devices()
        self.use_async = "nvlink" in strategy_name.lower()

        # Parse FGP allocations from context
        # Format: ("fgp:cuda:0,cuda:1", {"block.0.weight": "cuda:0", ...})
        self.fgp_components: Dict[str, Dict[str, str]] = {}
        self.component_devices: Dict[str, List[str]] = {}

        for comp_name, alloc in context.allocations.items():
            if isinstance(alloc, dict):
                device_str = alloc.get('device', '')
                shard_map = alloc.get('shard_map', {})
            elif isinstance(alloc, tuple) and len(alloc) >= 2:
                device_str, shard_map = alloc[0], alloc[1]
            else:
                continue

            if isinstance(device_str, str) and device_str.startswith("fgp:"):
                # This is an FGP-distributed component
                device_list = device_str[4:].split(",")  # "fgp:cuda:0,cuda:1" -> ["cuda:0", "cuda:1"]
                self.fgp_components[comp_name] = shard_map
                self.component_devices[comp_name] = device_list
            else:
                # Regular single-device component
                if isinstance(device_str, str):
                    self.component_devices[comp_name] = [device_str]

        # Primary device for non-FGP operations
        self.primary_device = self.devices[0] if self.devices else "cuda:0"

        # Track loaded components
        self._loaded_components: Set[str] = set()

        pass

    def is_fgp_component(self, component_name: str) -> bool:
        """Check if component is FGP-distributed."""
        return component_name in self.fgp_components

    def get_shard_map(self, component_name: str) -> Dict[str, str]:
        """Get shard map for FGP component."""
        return self.fgp_components.get(component_name, {})

    def get_component_devices(self, component_name: str) -> List[str]:
        """Get list of devices for component."""
        return self.component_devices.get(component_name, [self.primary_device])

    def execute_component(
        self,
        component_name: str,
        phase: str,
        inputs: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Execute component in FGP mode.

        For FGP components:
        - Enable FGP mode on executor (device alignment per op)
        - Load weights with shard map (direct to GPUs)
        - Execute graph with runtime device alignment
        """
        executor = self.context.component_executors.get(component_name)
        if executor is None:
            raise RuntimeError(
                f"ZERO FALLBACK: No executor for '{component_name}'"
            )

        is_fgp = self.is_fgp_component(component_name)
        devices = self.get_component_devices(component_name)
        device_str = ",".join(devices) if len(devices) > 1 else devices[0]

        # Load weights if not already loaded
        if component_name not in self._loaded_components:
            self._load_fgp_weights(component_name)
            self._loaded_components.add(component_name)

        # Set FGP mode on executor for proper device alignment
        # Note: _is_fgp_mode attribute may not exist on all executor types
        if is_fgp:
            if hasattr(executor, '_is_fgp_mode'):
                executor._is_fgp_mode = True  # type: ignore[attr-defined]
                pass

        # Prepare inputs
        if inputs:
            inputs = self.prepare_inputs(component_name, inputs)

        # Execute
        outputs = executor.run(inputs or {})

        # Disable FGP mode after execution if needed
        # (keep enabled for iterative components that run multiple times)

        return outputs

    def _load_fgp_weights(self, component_name: str) -> None:
        """
        Load weights for FGP component using shard map.

        Delegates to executor's load_weights with shard_map parameter
        which triggers WeightLoader._load_component_fgp().
        """
        executor = self.context.component_executors.get(component_name)
        if executor is None:
            return

        # Check if weights already loaded
        weights_loaded = hasattr(executor, '_weights') and bool(executor._weights)
        if weights_loaded:
            return

        shard_map = self.get_shard_map(component_name)
        devices = self.get_component_devices(component_name)

        # Get nbx_path from context
        nbx_path = None
        if hasattr(self.context, 'runtime_package') and self.context.runtime_package:
            nbx_path = getattr(self.context.runtime_package, 'nbx_path', None)

        if nbx_path:
            if shard_map:
                # The shard_map is passed through allocation and picked up by WeightLoader
                executor.load_weights(nbx_path, component_name, shard_map)
            else:
                executor.load_weights(nbx_path, component_name)

    def prepare_inputs(
        self,
        component_name: str,
        inputs: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Prepare inputs for FGP execution.

        For FGP components, inputs go to the first device in the device list.
        The graph executor handles per-op device alignment.
        """
        devices = self.get_component_devices(component_name)
        target_device = devices[0] if devices else self.primary_device

        return self.transfer_dict(inputs, target_device, self.use_async)

    def handle_outputs(
        self,
        component_name: str,
        outputs: Dict[str, Any],
        target_device: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Handle outputs from FGP component.

        Outputs may be on the last device in the FGP chain.
        Transfer to target_device if specified.
        """
        if target_device:
            # Find where outputs currently are
            source_device = None
            for key, val in outputs.items():
                if isinstance(val, torch.Tensor):
                    source_device = str(val.device)
                    break

            if source_device and source_device != target_device:
                outputs = self.transfer_dict(outputs, target_device, self.use_async)
                if self.use_async:
                    self.synchronize_device(target_device)

        return outputs

    def unload_weights(self, component_name: str) -> None:
        """
        Unload weights for FGP component.

        Clears weights from all devices where they were loaded.
        """
        executor = self.context.component_executors.get(component_name)
        if executor is not None:
            # Check if weights loaded
            weights_loaded = hasattr(executor, '_weights') and bool(executor._weights)
            if weights_loaded:
                devices = self.get_component_devices(component_name)
                # Disable FGP mode before unload
                if hasattr(executor, '_is_fgp_mode'):
                    executor._is_fgp_mode = False  # type: ignore[attr-defined]

                executor.unload_weights()
                self._loaded_components.discard(component_name)

                # Synchronize and clear cache on all devices
                for device in devices:
                    self.synchronize_device(device)

                torch.cuda.empty_cache()

    def get_embedding_device(self, component_name: str) -> str:
        """
        Get device where embedding layer is located.

        For FGP components, embeddings are typically on the first device.
        This is important for LLM token embedding lookups.
        """
        devices = self.get_component_devices(component_name)
        return devices[0] if devices else self.primary_device

    def get_head_device(self, component_name: str) -> str:
        """
        Get device where output head is located.

        For FGP components, the output head is typically on the last device.
        This is important for LLM logit computation.
        """
        devices = self.get_component_devices(component_name)
        return devices[-1] if devices else self.primary_device

    def transfer_to_embedding_device(
        self,
        tensor: torch.Tensor,
        component_name: str,
    ) -> torch.Tensor:
        """Transfer tensor to embedding device for component."""
        target = self.get_embedding_device(component_name)
        return self.transfer_tensor(tensor, target, self.use_async)

    def transfer_to_head_device(
        self,
        tensor: torch.Tensor,
        component_name: str,
    ) -> torch.Tensor:
        """Transfer tensor to head device for component."""
        target = self.get_head_device(component_name)
        return self.transfer_tensor(tensor, target, self.use_async)

    def synchronize_all(self) -> None:
        """Synchronize all devices used in FGP execution."""
        for device in self.devices:
            self.synchronize_device(device)


class FGPNVLinkStrategy(FGPStrategy):
    """FGP with NVLink high-speed transfers (default)."""

    def __init__(self, context: StrategyContext, strategy_name: str = "fgp_nvlink"):
        super().__init__(context, strategy_name)
        self.use_async = True


class FGPPCIeStrategy(FGPStrategy):
    """FGP with PCIe transfers (slower, for systems without NVLink)."""

    def __init__(self, context: StrategyContext, strategy_name: str = "fgp_pcie"):
        super().__init__(context, strategy_name)
        self.use_async = False



"""
Kernels for layer normalization.
"""


import triton
import triton.language as tl
from triton import next_power_of_2

from .softmax_kernels import BLOCK_SIZE_BATCH_heuristic
from .utils import warps_kernel_configs


@triton.autotune(
    configs=warps_kernel_configs(),
    key=['batch_dim', 'feat_dim'],
)
@triton.heuristics({'BLOCK_SIZE_BATCH': BLOCK_SIZE_BATCH_heuristic,
                    'BLOCK_SIZE_FEAT': lambda args: next_power_of_2(args['feat_dim'])})
@triton.jit
def layer_norm_forward_kernel(
    input_pointer, weight_pointer, bias_pointer,
    mean_pointer, inv_std_pointer, output_pointer,
    batch_dim, feat_dim,
    input_batch_stride, input_feat_stride,
    output_batch_stride, output_feat_stride,
    eps,
    scale_by_weight: tl.constexpr, add_bias: tl.constexpr, save_stats: tl.constexpr,
    BLOCK_SIZE_BATCH: tl.constexpr, BLOCK_SIZE_FEAT: tl.constexpr,
    ):
    """
    Layer-normalizes the input.

    Args:
        input_pointer: Pointer to the input to layer-normalize.
            The input must be of shape [batch_dim, feat_dim].
        weight_pointer: Pointer to optional weights for affine transform.
            The weights, if provided, must be of shape [feat_dim].
        bias_pointer: Pointer to an optional bias vector for affine transform.
            The bias vector, if provided, must be of shape [feat_dim].
        mean_pointer: Pointer to an optional container the input's mean
            is written to if save_stats is True.
            The container, if provided, must be of shape [batch_dim].
        inv_std_pointer: Pointer to an optional container the input's inverse
            standard deviation is written to if save_stats is True.
            The container, if provided, must be of shape [batch_dim].
        output_pointer: Pointer to a container the result is written to.
            The container must be of shape [batch_dim, feat_dim].
        batch_dim: Batch dimension.
        feat_dim: Dimensionality of the features.
        input_batch_stride: Stride necessary to jump one element along the
            input's batch dimension.
        input_feat_stride: Stride necessary to jump one element along the
            input's feature dimension.
        output_batch_stride: Stride necessary to jump one element along the
            output container's batch dimension.
        output_feat_stride: Stride necessary to jump one element along the
            output container's feature dimension.
        eps: Epsilon added in the square root in the denominator
            to avoid division by zero.
        scale_by_weight: Flag for scaling the normalized output by weights.
        add_bias: Flag for adding a bias vector to the normalized output
            if scale_by_weight is True.
        save_stats: Flag for saving the mean and standard deviation.
        BLOCK_SIZE_BATCH: Block size across the batch dimension.
        BLOCK_SIZE_FEAT: Block size across the feature dimension.
    """
    # This program processes BLOCK_SIZE_BATCH rows and BLOCK_SIZE_FEAT columns.
    batch_pid = tl.program_id(axis=0)

    batch_offset = batch_pid * BLOCK_SIZE_BATCH + tl.arange(0, BLOCK_SIZE_BATCH)
    feat_offset = tl.arange(0, BLOCK_SIZE_FEAT)

    batch_mask = batch_offset < batch_dim
    feat_mask = feat_offset < feat_dim

    input_pointer += (input_batch_stride * batch_offset[:, None] +
                      input_feat_stride * feat_offset[None, :])
    output_pointer += (output_batch_stride * batch_offset[:, None] +
                       output_feat_stride * feat_offset[None, :])

    input = tl.load(input_pointer,
                    mask=batch_mask[:, None] & feat_mask[None, :]).to(tl.float32)
    mean = tl.sum(input, axis=1) / feat_dim
    diff = tl.where(feat_mask[None, :], input - mean[:, None], 0)
    inv_std = tl.rsqrt(tl.sum(diff * diff, axis=1) / feat_dim + eps)

    if save_stats:
        tl.store(mean_pointer + batch_offset, mean, mask=batch_mask)
        tl.store(inv_std_pointer + batch_offset, inv_std, mask=batch_mask)

    output = diff * inv_std[:, None]
    if scale_by_weight:
        weight = tl.load(weight_pointer + feat_offset, mask=feat_mask)
        output *= weight
        if add_bias:
            bias = tl.load(bias_pointer + feat_offset, mask=feat_mask)
            output += bias

    tl.store(output_pointer, output,
             mask=batch_mask[:, None] & feat_mask[None, :])


@triton.autotune(
    configs=warps_kernel_configs(),
    key=['batch_dim', 'feat_dim'],
)
@triton.heuristics({'BLOCK_SIZE_BATCH': BLOCK_SIZE_BATCH_heuristic,
                    'BLOCK_SIZE_FEAT': lambda args: next_power_of_2(args['feat_dim'])})
@triton.jit
def layer_norm_backward_kernel(
    output_grad_pointer, input_pointer, mean_pointer, inv_std_pointer, weight_pointer,
    input_grad_pointer, weight_grad_pointer, bias_grad_pointer,
    batch_dim, feat_dim,
    output_grad_batch_stride, output_grad_feat_stride,
    input_batch_stride, input_feat_stride,
    input_grad_batch_stride, input_grad_feat_stride,
    weight_grad_batch_stride, weight_grad_feat_stride,
    bias_grad_batch_stride, bias_grad_feat_stride,
    scale_by_weight: tl.constexpr, add_bias: tl.constexpr,
    BLOCK_SIZE_BATCH: tl.constexpr, BLOCK_SIZE_FEAT: tl.constexpr,
    ):
    """
    Calculates the input gradient of layer normalization.

    Args:
        output_grad_pointer: Pointer to layer normalization's output gradients.
            The output gradients must be of shape [batch_dim, feat_dim].
        input_pointer: Pointer to the input.
            The input must be of shape [batch_dim, feat_dim].
        mean_pointer: Pointer to the input's mean.
            The mean should be of shape [batch_dim].
        inv_std_pointer: Pointer to the input's inverse standard deviation.
            The inverse standard deviation should be of shape [batch_dim].
        weight_pointer: Pointer to optional weights if affine transform occurred.
            The weights, if provided, must be of shape [feat_dim].
        input_grad_pointer: Pointer to a container the input's gradients are written to.
            The container must be of shape [batch_dim, feat_dim].
        weight_grad_pointer: Pointer to an optional container the weights' row-wise gradients
            are written to if scale_by_weight is True, which should later be summed.
            The container, if provided, must be of shape [batch_dim/BLOCK_SIZE_BATCH, feat_dim].
        bias_grad_pointer: Pointer to an optional container the bias vector's row-wise gradients
            are written to if scale_by_weight and add_bias are True, which should later be summed.
            The container, if provided, must be of shape [batch_dim/BLOCK_SIZE_BATCH, feat_dim].
        batch_dim: Batch dimension.
        feat_dim: Dimensionality of the features.
        output_grad_batch_stride: Stride necessary to jump one element along the
            output gradients' batch dimension.
        output_grad_feat_stride: Stride necessary to jump one element along the
            output gradients' feature dimension.
        input_batch_stride: Stride necessary to jump one element along the
            input's batch dimension.
        input_feat_stride: Stride necessary to jump one element along the
            input's feature dimension.
        input_grad_batch_stride: Stride necessary to jump one element along the
            input gradient container's batch dimension.
        input_grad_feat_stride: Stride necessary to jump one element along the
            input gradient container's feature dimension.
        weight_grad_batch_stride: Stride necessary to jump one element along the
            weight gradient container's batch dimension.
        weight_grad_feat_stride: Stride necessary to jump one element along the
            weight gradient container's feature dimension.
        bias_grad_batch_stride: Stride necessary to jump one element along the
            weight gradient container's batch dimension.
        bias_grad_feat_stride: Stride necessary to jump one element along the
            weight gradient container's feature dimension.
        scale_by_weight: Flag for scaling the normalized output by weights.
        add_bias: Flag for adding a bias vector to the normalized output
            if scale_by_weight is True.
        BLOCK_SIZE_BATCH: Block size across the batch dimension.
        BLOCK_SIZE_FEAT: Block size across the feature dimension.
    """
    # This program processes a single row and BLOCK_SIZE_FEAT columns.
    batch_pid = tl.program_id(axis=0)

    batch_offset = batch_pid * BLOCK_SIZE_BATCH + tl.arange(0, BLOCK_SIZE_BATCH)
    feat_offset = tl.arange(0, BLOCK_SIZE_FEAT)

    batch_mask = batch_offset < batch_dim
    feat_mask = feat_offset < feat_dim

    output_grad_pointer += (output_grad_batch_stride * batch_offset[:, None] +
                            output_grad_feat_stride * feat_offset[None, :])
    input_pointer += (input_batch_stride * batch_offset[:, None] +
                      input_feat_stride * feat_offset[None, :])
    input_grad_pointer += (input_grad_batch_stride * batch_offset[:, None] +
                           input_grad_feat_stride * feat_offset[None, :])

    output_grad = tl.load(output_grad_pointer,
                          mask=batch_mask[:, None] & feat_mask[None, :]).to(tl.float32)
    input = tl.load(input_pointer, mask=batch_mask[:, None] & feat_mask[None, :]).to(tl.float32)
    mean = tl.load(mean_pointer + batch_offset, mask=batch_mask)
    inv_std = tl.load(inv_std_pointer + batch_offset, mask=batch_mask)
    pre_lin = (input - mean[:, None]) * inv_std[:, None]

    if scale_by_weight:
        weight = tl.load(weight_pointer + feat_offset, mask=feat_mask)
        weight_output_grad_prod = weight * output_grad

    else:
        weight_output_grad_prod = output_grad

    term1 = tl.sum(pre_lin * weight_output_grad_prod, axis=1) / feat_dim
    term1 = pre_lin * term1[:, None]
    term2 = tl.sum(weight_output_grad_prod, axis=1) / feat_dim
    input_grad = (inv_std[:, None] *
                  (weight_output_grad_prod - (term1 + term2[:, None])))

    tl.store(input_grad_pointer, input_grad,
             mask=batch_mask[:, None] & feat_mask[None, :])

    if scale_by_weight:
        weight_grad_pointer += (weight_grad_batch_stride * batch_pid +
                                weight_grad_feat_stride * feat_offset)
        tl.store(weight_grad_pointer,
                 tl.sum(output_grad * pre_lin, axis=0),
                 mask=feat_mask)

        if add_bias:
            bias_grad_pointer += (bias_grad_batch_stride * batch_pid +
                                  bias_grad_feat_stride * feat_offset)
            tl.store(bias_grad_pointer,
                     tl.sum(output_grad, axis=0),
                     mask=feat_mask)



"""
Zero3 Strategy - CPU Offload with GPU Compute

Weights live on CPU (pinned memory), compute on GPU.
Double buffering with async prefetch for overlap.

ZERO HARDCODE: Block structure from topology, devices from Prism.
VENDORLESS: Uses device strings from Prism (cuda/hip/xpu).
ZERO IR: Uses TensorDAG directly.

Architecture (100% ATen/Triton):
- Uses GraphExecutor for TensorDAG execution
- Triton kernels for compute, PyTorch for metadata ops
"""

import re
import torch
from dataclasses import dataclass
from typing import Dict, List, Optional, Any, TYPE_CHECKING

from .base import ExecutionStrategy, StrategyContext

if TYPE_CHECKING:
    from neurobrix.core.prism.solver import ComponentAllocation


@dataclass
class BlockInfo:
    """Information about a model block."""
    index: int          # Sequential index (0, 1, 2, ...)
    block_id: int       # Original block ID from model
    weight_names: List[str]
    memory_mb: float


class Zero3Strategy(ExecutionStrategy):
    """
    CPU offload strategy with GPU compute.

    For models too large for GPU memory:
    - Weights stored on CPU (pinned memory)
    - Compute happens on GPU
    - Double buffering with async prefetch

    ZERO HARDCODE: Block structure from weight names.
    ZERO IR: Uses TensorDAG directly.
    VENDORLESS: Works with cuda/hip/xpu.
    """

    def __init__(self, context: StrategyContext, strategy_name: str = "zero3"):
        super().__init__(context, strategy_name)

        # Zero3 executor per component (lazy init)
        self._zero3_executors: Dict[str, 'Zero3Executor'] = {}

        # Get execution device from first component allocation
        self.exec_device = self._get_exec_device()

    def _get_exec_device(self) -> str:
        """Get execution device from allocations."""
        for comp_name, alloc_info in self.context.allocations.items():
            if isinstance(alloc_info, dict):
                device_str = alloc_info.get('device')
            else:
                # Type narrowing: if not dict, must be tuple
                assert isinstance(alloc_info, tuple), f"Allocation must be dict or tuple, got {type(alloc_info)}"
                device_str = alloc_info[0]

            if device_str and device_str != "cpu":
                if device_str.startswith("cuda"):
                    return device_str
        # Fallback to first available CUDA device
        if torch.cuda.is_available():
            return "cuda:0"
        return "cpu"

    def _ensure_executor(self, component_name: str) -> 'Zero3Executor':
        """Ensure Zero3Executor exists for component (lazy init)."""
        if component_name not in self._zero3_executors:
            executor = self.context.component_executors.get(component_name)
            if executor is None:
                raise RuntimeError(
                    f"ZERO FALLBACK: No executor found for component '{component_name}'. "
                    f"ExecutorFactory must create GraphExecutor first."
                )

            # Create Zero3Executor wrapping the GraphExecutor
            allocation = self._get_allocation(component_name)
            # Get DAG from GraphExecutor (attribute is _dag, not _graph)
            if not hasattr(executor, '_dag'):
                raise RuntimeError(
                    f"ZERO FALLBACK: Executor for '{component_name}' missing _dag attribute"
                )

            zero3_exec = Zero3Executor(
                component=component_name,
                executor=executor,
                exec_device=self.exec_device,
                dtype=executor.dtype,
            )
            self._zero3_executors[component_name] = zero3_exec

        return self._zero3_executors[component_name]

    def _get_allocation(self, component_name: str):
        """Get allocation for component."""
        return self.context.allocations.get(component_name)

    def execute_component(
        self,
        component_name: str,
        phase: str = "loop",
        inputs: Optional[Dict[str, torch.Tensor]] = None,
    ) -> Optional[Dict[str, Any]]:
        """Execute component with Zero3 CPU offload."""
        zero3_exec = self._ensure_executor(component_name)

        # Prepare inputs
        prepared_inputs = self.prepare_inputs(component_name, inputs or {})

        # Execute with Zero3
        outputs = zero3_exec.forward(prepared_inputs)

        return outputs

    def prepare_inputs(
        self,
        component_name: str,
        inputs: Dict[str, torch.Tensor],
    ) -> Dict[str, torch.Tensor]:
        """Transfer inputs to execution device."""
        return {
            name: tensor.to(self.exec_device) if hasattr(tensor, 'to') else tensor
            for name, tensor in inputs.items()
        }

    def handle_outputs(
        self,
        component_name: str,
        outputs: Dict[str, Any],
        target_device: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Outputs already on exec device."""
        if target_device and target_device != self.exec_device:
            return self.transfer_dict(outputs, target_device)
        return outputs

    def cleanup(self) -> None:
        """Release all Zero3 executor resources."""
        for executor in self._zero3_executors.values():
            executor.cleanup()
        self._zero3_executors.clear()


class Zero3Executor:
    """
    Execute model with ZeRO-3 style CPU offloading.

    - All weights on CPU (pinned memory for fast transfer)
    - Compute happens on GPU
    - Double buffering: prefetch block N+1 while computing block N
    - CUDA streams for async transfer

    ZERO HARDCODE: Block structure from weight names.
    ZERO IR: Uses TensorDAG directly.
    VENDORLESS: Works with cuda/hip/xpu.
    """

    def __init__(
        self,
        component: str,
        executor: Any,  # GraphExecutor
        exec_device: str,
        dtype: torch.dtype = torch.float16,
    ):
        """
        Initialize Zero3 executor.

        Args:
            component: Component name
            executor: GraphExecutor instance
            exec_device: GPU device for compute
            dtype: Weight dtype
        """
        self.component = component
        self.executor = executor
        self.exec_device = exec_device
        self._dtype = dtype

        # CPU weights storage (pinned for fast transfer)
        self.cpu_weights: Dict[str, torch.Tensor] = {}

        # GPU cache (double buffer)
        self.gpu_cache: List[Dict[str, torch.Tensor]] = [{}, {}]
        self.current_cache_idx = 0

        # CUDA stream for async transfer
        self.transfer_stream: Optional[torch.cuda.Stream] = None

        # Block organization
        self.blocks: List[BlockInfo] = []
        self.block_weights: Dict[int, Dict[str, torch.Tensor]] = {}
        self.num_blocks = 0

        # Initialize
        self._initialize()

    def _initialize(self) -> None:
        """Organize weights by block."""
        # Create transfer stream
        if torch.cuda.is_available() and self.exec_device.startswith("cuda"):
            device_idx = int(self.exec_device.split(":")[1])
            # Type narrowing: torch.cuda.Stream returns _CudaStreamBase which is compatible
            stream = torch.cuda.Stream(device=device_idx)
            self.transfer_stream = stream  # type: ignore[assignment]

        # Get weights from executor and pin to CPU
        if hasattr(self.executor, '_weights') and self.executor._weights:
            self._pin_weights_to_cpu(self.executor._weights)

        # Organize weights by block
        self._organize_by_block()

        pass

    def _pin_weights_to_cpu(self, weights: Dict[str, torch.Tensor]) -> None:
        """Pin weights to CPU memory for fast GPU transfer."""
        for name, tensor in weights.items():
            cpu_tensor = tensor.to("cpu")
            if cpu_tensor.is_contiguous():
                self.cpu_weights[name] = cpu_tensor.pin_memory()
            else:
                self.cpu_weights[name] = cpu_tensor.contiguous().pin_memory()

    def _organize_by_block(self) -> None:
        """
        Group weights by block index.

        Parse weight names to find block indices.
        Example: "transformer.blocks.5.attn.qkv.weight" -> block 5
        """
        block_weights: Dict[int, Dict[str, torch.Tensor]] = {}
        non_block_weights: Dict[str, torch.Tensor] = {}

        for name, tensor in self.cpu_weights.items():
            # Try to find block index in weight name
            block_idx = self._extract_block_index(name)

            if block_idx is not None:
                if block_idx not in block_weights:
                    block_weights[block_idx] = {}
                block_weights[block_idx][name] = tensor
            else:
                # Non-block weights (embeddings, final layers)
                non_block_weights[name] = tensor

        # Sort blocks
        sorted_blocks = sorted(block_weights.keys())

        # Distribute non-block weights to first and last blocks
        if sorted_blocks and non_block_weights:
            first_block = sorted_blocks[0]
            last_block = sorted_blocks[-1]

            for name, tensor in non_block_weights.items():
                name_lower = name.lower()
                # Embeddings/initial layers -> first block
                if any(hint in name_lower for hint in
                       ["embed", "patch", "pos", "input", "norm_pre", "time_"]):
                    block_weights[first_block][name] = tensor
                else:
                    # Final layers -> last block
                    block_weights[last_block][name] = tensor
        elif non_block_weights and not sorted_blocks:
            # No block structure, treat all as single block
            block_weights[0] = non_block_weights
            sorted_blocks = [0]

        self.block_weights = block_weights
        self.num_blocks = len(sorted_blocks) if sorted_blocks else 1

        # Create block info
        for idx, block_idx in enumerate(sorted_blocks):
            weights = block_weights[block_idx]
            memory_bytes = sum(t.numel() * t.element_size() for t in weights.values())
            memory_mb = memory_bytes / (1024 * 1024)
            self.blocks.append(BlockInfo(
                index=idx,
                block_id=block_idx,
                weight_names=list(weights.keys()),
                memory_mb=memory_mb,
            ))

        pass

    def _extract_block_index(self, weight_name: str) -> Optional[int]:
        """Extract block index from weight name."""
        patterns = [
            r'blocks?[._](\d+)',     # blocks.0, block_0
            r'layers?[._](\d+)',     # layer.0, layers_0
            r'transformer[._]h[._](\d+)',  # transformer.h.0 (GPT style)
            r'encoder[._]layer[._](\d+)',  # encoder.layer.0 (BERT style)
        ]
        for pattern in patterns:
            match = re.search(pattern, weight_name, re.IGNORECASE)
            if match:
                return int(match.group(1))
        return None

    def forward(
        self,
        inputs: Dict[str, torch.Tensor],
        sync: bool = True,
    ) -> Dict[str, torch.Tensor]:
        """
        Execute with block-by-block GPU transfer.

        Double buffering: prefetch next block while computing current.

        Args:
            inputs: Input tensors
            sync: Whether to synchronize at end

        Returns:
            Output tensors
        """
        # Move inputs to exec device
        hidden = {
            name: tensor.to(self.exec_device)
            for name, tensor in inputs.items()
        }

        for block_idx in range(self.num_blocks):
            # === PREFETCH NEXT BLOCK (async) ===
            if block_idx + 1 < self.num_blocks:
                self._prefetch_block(block_idx + 1)

            # === GET CURRENT BLOCK WEIGHTS ===
            block_weights = self._get_block_weights(block_idx)

            # === EXECUTE BLOCK ON GPU ===
            hidden = self._execute_block(hidden, block_weights, block_idx)

        # Final sync
        if sync and torch.cuda.is_available():
            torch.cuda.synchronize()

        return hidden

    def run(
        self,
        inputs: Dict[str, torch.Tensor],
        sync: bool = True,
    ) -> Dict[str, torch.Tensor]:
        """Alias for forward() for API compatibility."""
        return self.forward(inputs, sync=sync)

    def _prefetch_block(self, block_idx: int) -> None:
        """
        Async copy block weights from CPU to GPU.

        Uses separate CUDA stream for overlap with compute.
        """
        next_cache_idx = 1 - self.current_cache_idx
        block_info = self.blocks[block_idx]
        weights = self.block_weights[block_info.block_id]

        # Clear next cache
        self.gpu_cache[next_cache_idx].clear()

        # Async transfer
        if self.transfer_stream is not None:
            with torch.cuda.stream(self.transfer_stream):
                for name, cpu_tensor in weights.items():
                    self.gpu_cache[next_cache_idx][name] = cpu_tensor.to(
                        self.exec_device,
                        non_blocking=True,
                    )
        else:
            # Fallback: sync transfer
            for name, cpu_tensor in weights.items():
                self.gpu_cache[next_cache_idx][name] = cpu_tensor.to(
                    self.exec_device,
                )

    def _get_block_weights(self, block_idx: int) -> Dict[str, torch.Tensor]:
        """
        Get block weights from GPU cache.

        If first block, do sync copy. Otherwise, wait for prefetch.
        """
        if block_idx == 0:
            # First block: sync copy (no prefetch yet)
            block_info = self.blocks[0]
            weights = self.block_weights[block_info.block_id]

            for name, cpu_tensor in weights.items():
                self.gpu_cache[self.current_cache_idx][name] = cpu_tensor.to(
                    self.exec_device,
                )
        else:
            # Wait for prefetch to complete
            if self.transfer_stream is not None:
                self.transfer_stream.synchronize()
            # Switch cache
            self.current_cache_idx = 1 - self.current_cache_idx

        return self.gpu_cache[self.current_cache_idx]

    def _execute_block(
        self,
        hidden: Dict[str, torch.Tensor],
        weights: Dict[str, torch.Tensor],
        block_idx: int,
    ) -> Dict[str, torch.Tensor]:
        """
        Execute one block with weights on GPU.

        Temporarily inject weights into executor, run, then optionally clear.
        """
        # Inject block weights into executor
        self.executor._weights = weights

        # Execute through executor
        outputs = self.executor.run(hidden)

        return outputs

    def get_block_count(self) -> int:
        """Get number of blocks."""
        return self.num_blocks

    def get_total_cpu_memory_mb(self) -> float:
        """Get total CPU memory used for weights."""
        return sum(
            t.numel() * t.element_size()
            for t in self.cpu_weights.values()
        ) / (1024 * 1024)

    def cleanup(self) -> None:
        """Release all resources."""
        # Clear CPU weights
        self.cpu_weights.clear()

        # Clear GPU caches
        self.gpu_cache = [{}, {}]

        # Clear block weights
        self.block_weights.clear()
        self.blocks.clear()

        # Clear transfer stream
        self.transfer_stream = None

        # Clear GPU memory
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()
        return False



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang


import torch
import triton
import triton.language as tl

from fla.ops.utils.op import exp
from fla.utils import autotune_cache_kwargs, input_guard


@triton.heuristics({
    'USE_INITIAL_STATE': lambda args: args['h0'] is not None,
    'STORE_FINAL_STATE': lambda args: args['ht'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BD': BD}, num_warps=num_warps)
        for BD in [32, 64, 128]
        for num_warps in [1, 2, 4, 8]
    ],
    key=['D'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def fused_recurrent_hgrn_fwd_kernel(
    x,
    g,
    o,
    h0,
    ht,
    cu_seqlens,
    T,
    D: tl.constexpr,
    BD: tl.constexpr,
    USE_INITIAL_STATE: tl.constexpr,
    STORE_FINAL_STATE: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_d, i_n = tl.program_id(0), tl.program_id(1)
    if IS_VARLEN:
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int64), tl.load(cu_seqlens + i_n + 1).to(tl.int64)
        T = eos - bos
    else:
        bos, eos = i_n * T, i_n * T + T

    o_d = i_d * BD + tl.arange(0, BD)
    mask = o_d < D

    p_x = x + bos * D + o_d
    p_g = g + bos * D + o_d
    p_o = o + bos * D + o_d

    b_h = tl.zeros([BD], dtype=tl.float32)
    if USE_INITIAL_STATE:
        p_h0 = h0 + i_n * D + o_d
        b_h += tl.load(p_h0, mask=mask, other=0).to(tl.float32)
    for _ in range(0, T):
        b_x = tl.load(p_x, mask=mask, other=0).to(tl.float32)
        b_g = tl.load(p_g, mask=mask, other=0).to(tl.float32)
        b_h = exp(b_g) * b_h + b_x
        tl.store(p_o, b_h.to(p_o.dtype.element_ty), mask=mask)

        p_x += D
        p_g += D
        p_o += D

    if STORE_FINAL_STATE:
        p_ht = ht + i_n * D + o_d
        tl.store(p_ht, b_h.to(p_ht.dtype.element_ty), mask=mask)


@triton.heuristics({
    'USE_INITIAL_STATE': lambda args: args['h0'] is not None,
    'USE_FINAL_STATE_GRADIENT': lambda args: args['dht'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BD': BD}, num_warps=num_warps)
        for BD in [32, 64, 128]
        for num_warps in [1, 2, 4, 8]
    ],
    key=['D'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def fused_recurrent_hgrn_bwd_kernel(
    g,
    o,
    h0,
    dx,
    dg,
    do,
    dht,
    dh0,
    cu_seqlens,
    T,
    D: tl.constexpr,
    BD: tl.constexpr,
    USE_INITIAL_STATE: tl.constexpr,
    USE_FINAL_STATE_GRADIENT: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_d, i_n = tl.program_id(0), tl.program_id(1)
    if IS_VARLEN:
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int64), tl.load(cu_seqlens + i_n + 1).to(tl.int64)
        T = eos - bos
    else:
        bos, eos = i_n * T, i_n * T + T

    o_d = i_d * BD + tl.arange(0, BD)
    mask = o_d < D

    p_g = g + (bos + T - 1) * D + o_d
    p_o = o + (bos + T - 2) * D + o_d
    p_dx = dx + (bos + T - 1) * D + o_d
    p_dg = dg + (bos + T - 1) * D + o_d
    p_do = do + (bos + T - 1) * D + o_d

    b_dh = tl.zeros([BD], dtype=tl.float32)
    if USE_FINAL_STATE_GRADIENT:
        p_dht = dht + i_n * D + o_d
        b_dh += tl.load(p_dht, mask=mask, other=0).to(tl.float32)

    for i in range(T - 1, -1, -1):
        b_g = tl.load(p_g, mask=mask, other=0).to(tl.float32)
        b_do = tl.load(p_do, mask=mask, other=0).to(tl.float32)
        if i > 0:
            b_o = tl.load(p_o, mask=mask, other=0).to(tl.float32)
        elif USE_INITIAL_STATE:
            b_o = tl.load(h0 + i_n * D + o_d, mask=mask, other=0).to(tl.float32)
        else:
            b_o = tl.zeros([BD], dtype=tl.float32)

        b_dh = b_dh + b_do
        b_dx = b_dh
        b_dh = b_dh * exp(b_g)
        b_dg = b_dh * b_o
        tl.store(p_dx, b_dx.to(p_dx.dtype.element_ty), mask=mask)
        tl.store(p_dg, b_dg.to(p_dg.dtype.element_ty), mask=mask)

        p_g -= D
        p_o -= D
        p_dx -= D
        p_dg -= D
        p_do -= D

    if USE_INITIAL_STATE:
        p_dh0 = dh0 + i_n * D + o_d
        tl.store(p_dh0, b_dh.to(p_dh0.dtype.element_ty), mask=mask)


def fused_recurrent_hgrn_fwd(
    x: torch.Tensor,
    g: torch.Tensor,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False,
    cu_seqlens: torch.LongTensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    B, T, D = x.shape
    N = B if cu_seqlens is None else len(cu_seqlens) - 1

    o = torch.empty_like(x)
    final_state = x.new_empty(N, D) if output_final_state else None

    def grid(meta): return (triton.cdiv(D, meta['BD']), N)
    fused_recurrent_hgrn_fwd_kernel[grid](
        x=x,
        g=g,
        o=o,
        h0=initial_state,
        ht=final_state,
        cu_seqlens=cu_seqlens,
        T=T,
        D=D,
    )
    return o, final_state


def fused_recurrent_hgrn_bwd(
    g: torch.Tensor,
    o: torch.Tensor,
    do: torch.Tensor,
    dht: torch.Tensor = None,
    initial_state: torch.Tensor = None,
    cu_seqlens: torch.LongTensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    B, T, D = do.shape
    N = B if cu_seqlens is None else len(cu_seqlens) - 1

    dx = torch.empty_like(o, dtype=torch.float)
    dg = torch.empty_like(g, dtype=torch.float)
    dh0 = torch.empty_like(initial_state, dtype=torch.float) if initial_state is not None else None
    def grid(meta): return (triton.cdiv(D, meta['BD']), N)
    fused_recurrent_hgrn_bwd_kernel[grid](
        g=g,
        o=o,
        h0=initial_state,
        dx=dx,
        dg=dg,
        do=do,
        dht=dht,
        dh0=dh0,
        cu_seqlens=cu_seqlens,
        T=T,
        D=D,
    )
    return dx, dg, dh0


class FusedRecurrentHGRNFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    def forward(
        ctx,
        x: torch.Tensor,
        g: torch.Tensor,
        initial_state: torch.Tensor = None,
        output_final_state: bool = False,
        cu_seqlens: torch.LongTensor | None = None,
    ):
        o, ht = fused_recurrent_hgrn_fwd(
            x=x,
            g=g,
            initial_state=initial_state,
            output_final_state=output_final_state,
            cu_seqlens=cu_seqlens,
        )
        ctx.save_for_backward(g, o, initial_state)
        ctx.cu_seqlens = cu_seqlens
        return o, ht

    @staticmethod
    @input_guard
    def backward(ctx, do, dht=None):
        g, o, initial_state = ctx.saved_tensors
        cu_seqlens = ctx.cu_seqlens

        dx, dg, dh0 = fused_recurrent_hgrn_bwd(
            g=g,
            o=o,
            do=do,
            dht=dht,
            initial_state=initial_state,
            cu_seqlens=cu_seqlens,
        )
        return dx, dg, dh0, None, None


@torch.compiler.disable
def fused_recurrent_hgrn(
    x: torch.Tensor,
    g: torch.Tensor,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False,
    cu_seqlens: torch.LongTensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    r"""
    Args:
        x (torch.Tensor):
            inputs of shape `[B, T, D].
        g (torch.Tensor):
            Forget gates of shape `[B, T, D]`.
        initial_state (Optional[torch.Tensor]):
            Initial state of shape `[N, D]` for `N` input sequences.
            For equal-length input sequences, `N` equals the batch size `B`.
            Default: `None`.
        output_final_state (Optional[bool]):
            Whether to output the final state of shape `[N, D]`. Default: `False`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, D]`.
        final_state (torch.Tensor):
            Final state of shape `[N, D]` if `output_final_state=True` else `None`.

    Examples::
        >>> import torch
        >>> import torch.nn.functional as F
        >>> from einops import rearrange
        >>> from fla.ops.hgrn import fused_recurrent_hgrn
        # inputs with equal lengths
        >>> B, T, D = 4, 2048, 512
        >>> x = torch.randn(B, T, D, device='cuda')
        >>> g = F.logsigmoid(torch.randn(B, T, D, device='cuda'))
        >>> h0 = torch.randn(B, D, device='cuda')
        >>> o, ht = fused_recurrent_hgrn(x, g, initial_state=h0, output_final_state=True)
        # for variable-length inputs, the batch size `B` is expected to be 1 and `cu_seqlens` is required
        >>> x, g = map(lambda x: rearrange(x, 'b t d -> 1 (b t) d'), (x, g))
        # for a batch with 4 sequences, `cu_seqlens` with 5 start/end positions are expected
        >>> cu_seqlens = x.new_tensor([0, 2048, 4096, 6144, 8192], dtype=torch.long)
        >>> o_var, ht_var = fused_recurrent_hgrn(x, g, initial_state=h0, output_final_state=True, cu_seqlens=cu_seqlens)
        >>> assert o.allclose(o_var.view(o.shape))
        >>> assert ht.allclose(ht_var)
    """
    return FusedRecurrentHGRNFunction.apply(
        x,
        g,
        initial_state,
        output_final_state,
        cu_seqlens,
    )



#  Copyright (c) NXAI GmbH.
#  This software may be used and distributed according to the terms of the NXAI Community License Agreement.

import torch
import torch.nn.functional as F
from fla.ops.gla.recurrent_fuse import fused_recurrent_gla


def ceildiv(a, b):
    return -(a // -b)


def naive_recurrent_gla(
    q, k, v, gk, initial_state=None, output_final_state=False, causal=True
):
    orig_dtype = q.dtype
    q, k, v, gk = map(lambda x: x.float(), (q, k, v, gk))
    batch_size, n_heads, seq_len, d_head_k = q.shape
    _, _, _, d_head_v = v.shape
    h = torch.zeros(
        batch_size, n_heads, d_head_k, d_head_v, dtype=torch.float32, device=q.device
    )
    o = torch.zeros_like(v)
    scale = d_head_k**-0.5

    if initial_state is not None:
        h += initial_state

    for i in range(seq_len):
        q_i = q[:, :, i, :] * scale
        k_i = k[:, :, i]
        v_i = v[:, :, i, :]
        gk_i = gk[:, :, i].exp()
        kv_i = k_i[..., None] * v_i[..., None, :]
        h = h * gk_i[..., None] + kv_i
        o_i = (q_i[..., None] * h).sum(-2)
        o[:, :, i] = o_i

    if causal:
        return o.to(orig_dtype), h
    else:
        o_reverse = torch.zeros_like(v)
        h = torch.zeros(
            batch_size,
            n_heads,
            d_head_k,
            d_head_v,
            dtype=torch.float32,
            device=q.device,
        )
        for i in range(seq_len - 1, -1, -1):
            q_i = q[:, :, i, :] * scale
            k_i = k[:, :, i]
            v_i = v[:, :, i, :]
            gk_i = gk[:, :, i].exp()
            kv_i = k_i[..., None] * v_i[..., None, :]
            h = h * gk_i[..., None] + kv_i
            o_i = (q_i[..., None] * h).sum(-2)
            o_reverse[:, :, i] = o_i

        return o, o_reverse


if __name__ == "__main__":
    B = 4
    H = 4
    L = 512
    D = 128
    dtype = torch.float32
    q = (torch.randn(B, H, L, D).cuda().to(dtype)).requires_grad_(True)
    k = (torch.randn(B, H, L, D).cuda().to(dtype)).requires_grad_(True)
    v = torch.randn(B, H, L, D).cuda().to(dtype).requires_grad_(True)
    g = (
        F.logsigmoid(torch.rand(B, H, L, D))
        .cuda()
        .clamp_min(-1)
        .to(torch.float32)
        .requires_grad_(True)
    )

    do = torch.rand_like(v).cuda()
    do2 = torch.rand_like(v).cuda()
    intial_state = torch.rand(B, H, D, D).cuda()

    ref, ref_rev = naive_recurrent_gla(q, k, v, g, causal=False)

    ref.backward(do, retain_graph=True)
    ref_rev.backward(do2, retain_graph=True)

    ref_dq, q.grad = q.grad.clone(), None
    ref_dk, k.grad = k.grad.clone(), None
    ref_dv, v.grad = v.grad.clone(), None
    ref_dg, g.grad = g.grad.clone(), None

    tri, tri_rev = fused_recurrent_gla(
        q,
        k,
        v,
        g,
        initial_state=None,
        scale=D**-0.5,
        output_final_state=False,
        causal=False,
    )
    tri.backward(do, retain_graph=True)
    tri_rev.backward(do2, retain_graph=True)
    tri_dq, q.grad = q.grad.clone(), None
    tri_dk, k.grad = k.grad.clone(), None
    tri_dv, v.grad = v.grad.clone(), None
    tri_dg, g.grad = g.grad.clone(), None

    assert ref.allclose(tri, 0, 1e-5)  # , breakpoint()
    assert ref_rev.allclose(tri_rev, 0, 1e-5)  # , breakpoint()
    assert ref_dq.allclose(tri_dq, 0, 1e-5)  # , breakpoint()
    assert ref_dk.allclose(tri_dk, 0, 1e-5)  # , breakpoint()
    assert ref_dv.allclose(tri_dv, 0, 1e-5)  # , breakpoint()
    assert ref_dg.allclose(tri_dg, 0, 1e-4)  # , breakpoint()

    # tri = fused_chunk_gla(q, k, v, g)
    # tri.backward(do, retain_graph=True)
    # tri_dq, q.grad = q.grad.clone(), None
    # tri_dk, k.grad = k.grad.clone(), None
    # tri_dv, v.grad = v.grad.clone(), None
    # tri_dg, g.grad = g.grad.clone(), None

    # assert ref.allclose(tri, 0, 1e-5), breakpoint()
    # assert ref_dq.allclose(tri_dq, 0, 1e-5), breakpoint()
    # assert ref_dk.allclose(tri_dk, 0, 1e-5), breakpoint()
    # assert ref_dv.allclose(tri_dv, 0, 1e-5), breakpoint()
    # assert ref_dg.allclose(tri_dg, 0, 1e-4), breakpoint()
    # breakpoint()
    print("Pass")
